/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file ftallocators.c
 * @brief Allocator procedures
 */
#include "ftalloc.h"
#include "ftlibtypes.h"

static int offsets[ftdev_max];

#define FTOPI_ALLOC(cmagic, num, mode, code) \
		ftobjInput *result = NULL; \
	    /* Only I1-I8 can be used */ \
	    if (num < ftdev_i1 || num > ftdev_i8) \
	        return result; \
	    int i = num - ftdev_i1; \
	    result = &txt2->inputs[i]; \
	    result->object.magic = cmagic; \
	    /* Fix configuration */ \
	    txt2->connusage |= 1 << num; \
	    txt2->new_config.inputs.inputs[i][0] = mode; \
	    txt2->new_config.inputs.inputs[i][1] = code; \
	    txt2->configid++; \
	    return result;

/**
 @brief Init object structures
 @param txt TXT2 controller
 */
void ftop_init_offsets(ftobject_txt2 *txt)
{
    int i;

    offsets[ftdev_txt] = 0;

    /* Process inputs */
    for (i = 0; i < FT_MAX_INPUTS; i++) {
        int j = i + ftdev_i1;
        offsets[j] = (char *)&txt->inputs[i] - (char *)txt;
        txt->inputs[i].connection.conn = j;
    }

    /* Process outputs */
    for (i = 0; i < FT_MAX_OUTPUTS; i++) {
        int j = i + ftdev_o1;
        offsets[j] = (char *)&txt->outputs[i] - (char *)txt;
        txt->outputs[i].connection.conn = j;
    }

    /* Process counters */
    for (i = 0; i < FT_MAX_COUNTERS; i++) {
        int j = i + ftdev_c1;
        offsets[j] = (char *)&txt->counters[i] - (char *)txt;
        txt->counters[i].connection.conn = j;
    }

    /* Process servo */
    for (i = 0; i < FT_MAX_SERVO; i++) {
        int j = i + ftdev_s1;
        offsets[j] = (char *)&txt->servo[i] - (char *)txt;
        txt->servo[i].connection.conn = j;
    }
}

/**
 @brief Get TXT2 for device
 @param [in] device device connected to TXT2 controller
 @return TXT2 device
 */
void *ftop_get_txt(void *device)
{
    ftobject_connection *obj = device;

    return (char*)obj - offsets[obj->conn];
}

/**
 @brief Set input extended range
 @param [in] device device
 @param [in] mode new mode: 0 - standard range, 1 - extended range
 */
void ftop_extend_range(void *device, int mode)
{
	ftobject_connection* object = device;
	int num = object->conn;
	ftobject_txt2 *txt2 = (void *) ((char *) device - offsets[num]);
	int i = num - ftdev_i1;

	if (mode) {
		txt2->new_config.inputs.inputs[i][0] |= ftinp_r2;
	} else {
		txt2->new_config.inputs.inputs[i][0] &= ~ftinp_r2;
	}
}

/**
 @brief Allocate resources for resistor
 @param [in] txt txt controller
 @param [in] num connection number
 @return pointer to resistor structure
 */
void *ftop_alloc_resistor(void *txt, int num)
{
    ftobject_txt2 *txt2 = txt;

    FTOPI_ALLOC(ft_resistor_magic, num, ftinp_passive | ftinp_analog, ftinput_resistor);
}

/**
 @brief Allocate resources for NCTTermistor
 @param [in] txt txt controller
 @param [in] num connection number
 @return pointer to NCTTermistor structure
 */
void *ftop_alloc_ncttermistor(void *txt, int num)
{
    ftobject_txt2 *txt2 = txt;

    FTOPI_ALLOC(ft_termistor_magic, num, ftinp_passive | ftinp_analog, ftinput_termistor);
}

/**
 @brief Allocate resources for switch
 @param [in] txt txt controller
 @param [in] num connection number
 @return pointer to switch structure
 */
void *ftop_alloc_switch(void *txt, int num)
{
    ftobject_txt2 *txt2 = txt;

    FTOPI_ALLOC(ft_switch_magic, num, ftinp_passive | ftinp_digital, ftinput_switch);
}

/**
 @brief Allocate resources for digital
 @param [in] txt txt controller
 @param [in] num connection number
 @return pointer to digital structure
 */
void *ftop_alloc_digital(void *txt, int num)
{
    ftobject_txt2 *txt2 = txt;

    FTOPI_ALLOC(ft_digital_magic, num, ftinp_active | ftinp_digital, ftinput_digital);
}

/**
 @brief Allocate resources for Voltmeter
 @param [in] txt txt controller
 @param [in] num connection number
 @return pointer to Voltmeter structure
 */
void *ftop_alloc_voltmeter(void *txt, int num)
{
    ftobject_txt2 *txt2 = txt;

    FTOPI_ALLOC(ft_voltmeter_magic, num, ftinp_active | ftinp_analog, ftinput_voltage);
}

/**
 @brief Allocate resources for ColorSensor
 @param [in] txt txt controller
 @param [in] num connection number
 @return pointer to ColorSensor structure
 */
void *ftop_alloc_colorsensor(void *txt, int num)
{
    ftobject_txt2 *txt2 = txt;

	FTOPI_ALLOC(ft_color_sensor_magic, num, ftinp_active | ftinp_analog,
			ftinput_color_sensor);
}

/**
 @brief Allocate resources for TrailFollower sensor
 @param [in] txt txt controller
 @param [in] num connection number
 @return pointer to TrailFollower structure
 */
void *ftop_alloc_trail_follower(void *txt, int num)
{
    ftobject_txt2 *txt2 = txt;

	FTOPI_ALLOC(ft_trail_follower_magic, num, ftinp_active | ftinp_digital,
			ftinput_trail_follower);
}

/**
 @brief Allocate resources for Ultrasonic sensor
 @param [in] txt txt controller
 @param [in] num connection number
 @return pointer to Ultrasonic structure
 */
void *ftop_alloc_ultrasonic(void *txt, int num)
{
    ftobject_txt2 *txt2 = txt;

    FTOPI_ALLOC(ft_ultrasonic_magic, num, ftinp_ultrasonic | ftinp_analog,
    		ftinput_ultrasonic);
}

/**
 @brief Allocate resources for Lamp
 @param [in] txt txt controller
 @param [in] num connection number
 @return pointer to Lamp structure
 */
void *ftop_alloc_lamp(void *txt, int num)
{
    ftobject_txt2 *txt2 = txt;
    ftobjOutput *result = NULL;
    int i;

    /* Only O1-O8 can be used */
    if (num < ftdev_o1 || num > ftdev_o8)
        return result;

    i = num - ftdev_o1;
    result = &txt2->outputs[i];
    result->object.magic = ft_lamp_magic;

    /* Fix configuration */
    txt2->connusage |= 1 << num;
    txt2->new_config.motors.usage[i/2] = ft_output_mode;
    txt2->configid++;

    return result;
}

/**
 @brief Allocate resources for Motor
 @param [in] txt txt controller
 @param [in] num connection number
 @return pointer to Motor structure
 */
void *ftop_alloc_motor(void *txt, int num)
{
    ftobject_txt2 *txt2 = txt;
    ftobjOutput *result = NULL;
    int i;

    /* Only O1, O3, O5, O7 can be used */
    if (num < ftdev_m1 || num > ftdev_m4 || ((num - ftdev_o1) & 1) )
        return result;

    i = num - ftdev_o1;
    result = &txt2->outputs[i];
    result->object.magic = ft_motor_magic;
    result[1].object.magic = ft_motor_magic;

    /* Fix configuration */
    txt2->connusage |= 3 << num;
    txt2->new_config.motors.usage[i/2] = ft_motor_mode;
    txt2->configid++;

    /* Set start cmdid */
    result->motor.cmdid = txt2->current_values.cmdvalues.cmdids[i/2];

    return result;
}

/**
 @brief Allocate resources for Encoder
 @param [in] txt txt controller
 @param [in] num connection number
 @return pointer to Encoder structure
 */
void *ftop_alloc_encoder(void *txt, int num)
{
    ftobject_txt2 *txt2 = txt;
    ftobjOutput *result = NULL;
    int i;

    /* Only O1, O3, O5, O7 can be used */
    if (num < ftdev_m1 || num > ftdev_m4 || ((num - ftdev_o1) & 1) )
        return result;

    i = num - ftdev_o1;
    result = &txt2->outputs[i];
    result->object.magic = ft_encoder_magic;
    result[1].object.magic = ft_encoder_magic;

    /* Fix configuration */
    txt2->connusage |= 3 << num;
    txt2->new_config.motors.usage[i/2] = ft_motor_mode;
    /* Set start cmdid */
    result->motor.cmdid = txt2->current_values.cmdvalues.cmdids[i/2];

    num = (num - ftdev_m1) / 2 + ftdev_c1;
    ftop_alloc_motor_step_counter(txt, num);

    return result;
}

/**
 @brief Allocate resources for MotorStepCounter
 @param [in] txt txt controller
 @param [in] num connection number
 @return pointer to MotorStepCounter structure
 */
void *ftop_alloc_motor_step_counter(void *txt, int num)
{
    ftobject_txt2 *txt2 = txt;
    ftobjCounter *result = NULL;
    int i;

    /* Only C1-C4 can be used ? */
    if (num < ftdev_c1 || num > ftdev_c4)
        return result;

    i = num - ftdev_c1;
    result = &txt2->counters[i];
    result->object.magic = ft_motor_step_counter_magic;

    /* Fix configuration */
    txt2->connusage |= 1 << num;
    txt2->configid++;

    return result;
}

/**
 @brief Allocate resources for Counter
 @param [in] txt txt controller
 @param [in] num connection number
 @return pointer to Counter structure
 */
void *ftop_alloc_counter(void *txt, int num)
{
    ftobject_txt2 *txt2 = txt;
    ftobjCounter *result = NULL;
    int i;

    /* Only C1-C4 can be used ? */
    if (num < ftdev_c1 || num > ftdev_c4)
        return result;

    i = num - ftdev_c1;
    result = &txt2->counters[i];
    result->object.magic = ft_counter_magic;

    /* Fix configuration */
    txt2->connusage |= 1 << num;
    txt2->configid++;

    return result;
}

/**
 @brief Allocate resources for Servo motor
 @param [in] txt txt controller
 @param [in] num connection number S1-S3
 @return pointer to Servo motor structure
 */
void *ftop_alloc_servo(void *txt, int num)
{
    ftobject_txt2 *txt2 = txt;
    ftobjServo *result = NULL;
    int i;

    /* Only S1-S3 can be used */
    if (num < ftdev_s1 || num > ftdev_s3)
        return result;

    i = num - ftdev_s1;
    result = &txt2->servo[i];
    result->object.magic = ft_servo_magic;

    /* Fix configuration */
    txt2->connusage |= 1 << num;
    txt2->configid++;

    return result;
}

/**
 @brief Free resources allocated for device
 @param [in] device allocated device
 */
void ftop_free_device(void *device)
{
	ftobject_connection* object = device;
	int num = object->conn;
	ftobject_txt2 *txt2 = (void *) ((char *) device - offsets[num]);

	/* Check motor case */
	if (object->magic == ft_motor_magic || object->magic == ft_encoder_magic) {
		ftobjOutput *fo;
		num &= ~1;
		fo = &txt2->outputs[num - ftdev_o1];
		fo->connection.magic = 0;
		fo[1].connection.magic = 0;
		txt2->connusage = ~(3 << num);
		/* TODO: counters ? */
	} else {
		txt2->connusage = ~(1 << num);
		object->magic = 0;
		if (num >= ftdev_i1 && num <= ftdev_i8) {
			/* Unconfigure */
			int i = num - ftdev_i1;
			txt2->new_config.inputs.inputs[i][0] = ftinp_unknown;
			txt2->new_config.inputs.inputs[i][1] = ftinput_unknown;
		}
	}
}

/**
 @brief Free resources allocated for device
 @param [in] device allocated device
 @return connection number
 */
int ftop_get_conn(ftdevice *device)
{
	ftobject_connection* object = device;

	return object->conn;
}

/**
 @brief Get allocation mode for connection
 @param [in] txt txt controller
 @param [in] num connection number
 @return allocation mode
 */
int ftop_alloc_mode(fttxt2 *txt, int num)
{
	ftobject_txt2 *txt2 = txt;
	struct ft_shm_txt *s = txt2->txt;
	int result = ftinp_unknown;

	if (num >= ftdev_i1 && num <= ftdev_i8) {
		/* Check local */
		int n = num - ftdev_i1;
		if (txt2->connusage & (1 << num))
			return txt2->new_config.inputs.inputs[n][0];
		/* Try to check shared memory */
		return s->config.inputs.inputs[n][0];
	} else if (num >= ftdev_o1 && num <= ftdev_o8) {
		int n = (num - ftdev_o1) / 2;
		if (txt2->connusage & (1 << num))
			return txt2->new_config.motors.usage[n];
		/* Try to check shared memory */
		return s->config.motors.usage[n];
	}
	return result;
}

/**
 @brief Get allocation code for connection
 @param [in] txt txt controller
 @param [in] num connection number
 @return allocation code
 */
int ftop_alloc_code(fttxt2 *txt, int num)
{
	ftobject_txt2 *txt2 = txt;
	struct ft_shm_txt *s = txt2->txt;
	int result = ftinput_unknown;

	if (num >= ftdev_i1 && num <= ftdev_i8) {
		/* Check local */
		int n = num - ftdev_i1;
		if (txt2->connusage & (1 << num))
			return txt2->new_config.inputs.inputs[n][1];
		/* Try to check shared memory */
		return s->config.inputs.inputs[n][1];
	} else if (num >= ftdev_o1 && num <= ftdev_o8) {
		int n = (num - ftdev_o1) / 2;
		if (txt2->connusage & (1 << num))
			return txt2->new_config.motors.usage[n];
		/* Try to check shared memory */
		return s->config.motors.usage[n];
	}
	return result;
}

/**
 @brief Check usage for connection
 @param [in] txt txt controller
 @param [in] num connection number
 @return usage
 */
int ftop_check_usage(fttxt2 *txt, int num)
{
	ftobject_txt2 *txt2 = txt;

	if (txt2->connusage & (1 << num))
		return 1;

	return 0;
}
