/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file ftslave.c
 * @brief Slave specific user level procedures
 *
 * TXT2 top slave procedures
 */

#include "ftslave.h"
#include "fttransport.h"
#include "ftlibtypes.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

extern void ftop_init_offsets(ftobject_txt2 *txt);

/**
 @brief Get TXT2 name
 @param [in] device TXT2 controller
 @return Device name
 */
const char* ftop_get_name(void *device)
{
    ftobject_txt2 *txt2 = device;

    return txt2->name;
}

/**
 @brief Get TXT2 version
 @param [in] device TXT2 controller
 @return Device version
 */
int ftop_get_version(void *device)
{
    ftobject_txt2 *txt2 = device;

    return txt2->version;
}

/**
 @brief Get TXT2 firmware version
 @param [in] device TXT2 controller
 @return Device firmware version
 */
int ftop_get_firmware(void *device)
{
    ftobject_txt2 *txt2 = device;
    return txt2->firmware;
}

/**
 @brief Get TXT2 uid
 @param [in] device TXT2 controller
 @return uid
 */
int ftop_get_uid(void *device)
{
    ftobject_txt2 *txt2 = device;

    return txt2->uid;
}

/**
 @brief Setup TXT2 mode
 @param [in] device TXT2 controller
 @param [in] mode master/slave mode
 @return error code
 */
int ftop_txt_mode(void *device, int mode)
{
    ftobject_txt2 *txt2 = device;

    if (txt2->canid) {
    	return 0;
    }

	struct ftop_set_mode set_mode = {
		{ sizeof(set_mode), ftop_op_set_mode, txt2->canid, txt2->cmdid++ }, mode };

	ftop_socket_send(txt2->transport, &set_mode, sizeof(set_mode));

    txt2->mode = mode;
    return 0;
}

/**
 @brief Setup TXT2 name
 @param [in] device TXT2 controller
 @param [in] name TXT2 name for discovery
 @return error code
 */
int ftop_txt_name(void *device, const char *name)
{
	ftobject_txt2 *txt2 = device;
	struct ftop_set_name set_name = {
		{ sizeof(set_name), ftop_op_set_name, txt2->canid, txt2->cmdid++ }, { 0 } };

	strncpy(set_name.name, name, sizeof(set_name.name));
	ftop_socket_send(txt2->transport, &set_name, sizeof(set_name));

	strncpy(txt2->name, name, sizeof(txt2->name));
	return 0;
}

/**
 @brief Get number of slaves TXT2
 @param [in] device TXT2 controller
 @return number of slaves TXT2
 */
unsigned int ftop_get_slaves_number(void *device){
    ftobject_txt2 *txt2 = device;

    if (txt2->canid)
    	return 0;

    ftuid *uids = txt2->suids;

	for (int i = 1; i < FT_MAX_TXT; i++) {
		if (uids[i] == 0)
			return i - 1;
	}

    return FT_MAX_TXT - 1; /* All slaves used */
}

/**
 @brief Get remote slave TXT2
 @param [in] txt2 master TXT2 controller
 @param [in] number slave number
 @return slave TXT2 structure
 */
static void *ftop_get_slave(ftobject_txt2 *txt2, unsigned int number)
{
	ftobject_txt2 *txt;
	struct ft_shm_userver *userver;

	userver = (void *) txt2->header;
	if (txt2->canid || userver == NULL || !number || number >= FT_MAX_TXT)
		return NULL;

	void* txts = (char *)userver + ((uint32_t)userver->header.txts - txt2->uaddr);
	uint32_t txtaddr = ((uint32_t *)txts)[number]; /* Internal address */
	struct ft_shm_txt *s = (void *) ((char *) userver + (txtaddr - txt2->uaddr));

	if (s->common.uid == 0)
		return NULL;

	txt = calloc(1, sizeof(txt[0]));
	txt->canid = number; /* TODO */
	txt->cvalues = &s->values;
	txt->firmware = s->common.firmware;
	txt->main = txt2;
	txt->mode = ft_txt_slave;
	strncpy(txt->name, s->common.device_name, sizeof(txt->name));
	txt->transport = txt2->transport;
	txt->txt = s;
	txt->uaddr = txt2->uaddr;
	txt->uid = s->common.uid;
	txt->version = s->common.version;
	txt2->slaves[number - 1].txt2 = txt;
	ftop_init_offsets(txt);

	return txt;
}

/**
 @brief Get TXT2 slave
 @param [in] device TXT2 controller
 @param [in] number slave number
 @return TXT2 device
 */
void *ftop_get_slave_by_order(void *device, unsigned int number)
{
    ftobject_txt2 *txt2 = device;

    if (txt2->canid || !number || number >= FT_MAX_TXT)
        return NULL;

    if (txt2->slaves[number - 1].txt2 == NULL)
    	return ftop_get_slave(txt2, number);

    return txt2->slaves[number - 1].txt2;
}

/**
 @brief Get TXT2 slave uid
 @param [in] device TXT2 controller
 @param [in] number slave number
 @return slave uid
 */
uint32_t ftop_get_txt_uid(void *device, unsigned int number)
{
	ftobject_txt2 *txt2 = device;
	ftuid *uids = txt2->suids;

	if (txt2->canid || number >= FT_MAX_TXT)
		return 0;

	return uids[number];
}

/**
 @brief Get TXT2 slave name
 @param [in] device master TXT2 controller
 @param [in] number slave number
 @return slave name
 */
const char *ftop_get_txt_name(void *device, unsigned int number)
{
	ftobject_txt2 *txt2 = device;

	if (txt2->canid || number >= FT_MAX_TXT)
		return NULL;

	struct ft_shm_userver *userver = (void *) txt2->header;
	if (userver == NULL)
		return NULL;

	void* txts = (char *)userver + ((uint32_t)userver->header.txts - txt2->uaddr);
	uint32_t txtaddr = ((uint32_t *)txts)[number]; /* Internal address */
	struct ft_shm_txt *s = (void *) ((char *) userver + (txtaddr - txt2->uaddr));

	if (s->common.device_name[0] == 0)
		return NULL;

	return s->common.device_name;
}

/**
 @brief Get TXT2 slave serial number
 @param [in] device master TXT2 controller
 @param [in] number slave number
 @return slave serial number
 */
const char *ftop_get_txt_SN(void *device, unsigned int number)
{
	ftobject_txt2 *txt2 = device;

	if (txt2->canid || number >= FT_MAX_TXT)
		return NULL;

	struct ft_shm_userver *userver = (void *) txt2->header;
	if (userver == NULL)
		return NULL;

	void* txts = (char *)userver + ((uint32_t)userver->header.txts - txt2->uaddr);
	uint32_t txtaddr = ((uint32_t *)txts)[number]; /* Internal address */
	struct ft_shm_txt *s = (void *) ((char *) userver + (txtaddr - txt2->uaddr));

	if (s->common.sn[0] == 0)
		return NULL;

	return s->common.sn;
}

/**
 @brief Get TXT2 slave
 @param [in] device master TXT2 controller
 @param [in] uid slave uid
 @return TXT2 device
 */
void *ftop_get_slave_by_uid(void *device, unsigned int uid)
{
	ftobject_txt2 *txt2 = device;

	if (txt2->canid)
		return NULL;

	for (unsigned int i = 0; i < FT_MAX_TXT; i++) {
		if (ftop_get_txt_uid(txt2, i) == uid) {
			if (!i)
				return device;

			if (txt2->slaves[i - 1].txt2 == NULL)
				return ftop_get_slave(txt2, i);

			return txt2->slaves[i - 1].txt2;
		}
	}

	return NULL;
}

/**
 @brief Get TXT2 slave
 @param [in] device master TXT2 controller
 @param [in] name slave name
 @return TXT2 device
 */
void *ftop_get_slave_by_name(void *device, const char *name)
{
	ftobject_txt2 *txt2 = device;

	if (txt2->canid)
		return NULL;

	for (unsigned int i = 0; i < FT_MAX_TXT; i++) {
		const char *sname = ftop_get_txt_name(txt2, i);

		if (sname != NULL && !strncmp(sname, name, FT_NAME_SIZE)) {
			if (!i)
				return device;

			if (txt2->slaves[i - 1].txt2 == NULL)
				return ftop_get_slave(txt2, i);

			return txt2->slaves[i - 1].txt2;
		}
	}

	return NULL;
}

/**
 @brief Get TXT2 slave
 @param [in] device master TXT2 controller
 @param [in] sn slave serial number
 @return TXT2 device
 */
void *ftop_get_slave_by_SN(void *device, const char *sn)
{
	ftobject_txt2 *txt2 = device;

	if (txt2->canid)
		return NULL;

	for (unsigned int i = 0; i < FT_MAX_TXT; i++) {
		const char *sname = ftop_get_txt_SN(txt2, i);
		if (sname != NULL && !strncmp(sname, sn, FT_NAME_SIZE)) {
			if (!i)
				return device;

			if (txt2->slaves[i - 1].txt2 == NULL)
				return ftop_get_slave(txt2, i);

			return txt2->slaves[i - 1].txt2;
		}
	}

	return NULL;
}
