/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file ftconfig.c
 * @brief TXT2 configuration utility
 */

#include "ftslave.h"
#include "ftop.h"
#include "ftlibtypes.h"
#include "ftcalib.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/rpmsg.h>
#include <signal.h>

#include <linux/i2c.h>
#include <linux/i2c-dev.h>

#define STIMEOUT      1 /*!< Timeout for Linux M4 stop/start commands */
#define MAX_ENDPOINTS 8 /*!< Maximum number of rpmsg endpoints - clients */
#define RPMSG_SERVICE_NAME "rpmsg-raw-channel"	/*!< M4 RPMSG service name */
#define RPMSG_DEV "/dev/rpmsg_ctrl0"			/*!< Standard RPMSG control device */
#define FT_CONF_FILE "/.ftconfig"				/*!< Standard configuration file name */
#define HW_CONFIG "/opt/ft/config/hwconfig"     /*!< Standard configuration path */
#define M4_MEM_SIZE (256 * 1024)				/*!< M4 memory size */
#define M4_ADDR 0x10000000						/*!< M4 firmware address */

static char *stateFile = "/sys/class/remoteproc/remoteproc0/state";
static char *firmwareFile = "/sys/class/remoteproc/remoteproc0/firmware";

static char *executable;
static int save = 0;
static char *conffile = NULL;
static int fdterm = 1; /* switch on */
static int role = 0;   /* single mode */
static unsigned int nslave = 0;  /* slave number undefined */
static unsigned int bitrate = 1000000; /* FD CAN bitrate */
static int climit = 2500; /* 2550 ma limit for 9v */
static uint32_t buid = 0; /* Board uid */
static uint32_t kuid = 0; /* Kontron uid */
static char tname[17];

static const char *roles[] = { "single", "master", "slave" };
static int verbose = 0;
static int loadcalib = 1;
static int startDaemon = 1;
static char *calib_file = "/opt/ft/config/ftcalib";
static char *cname;
static int rev01 = 0;

/**
 @brief Standard usage procedure
 */
void Usage()
{
	fprintf(stderr,
			"Usage: ftconfig [-c conf_file] [-s] [-v] [-e file] command params\n"
					"  -s           - update configuration file\n"
					"  -c conf_file - use new configuration file instead of .ftconfig\n"
					"  -e exe_file  - use executable file\n"
					"  -v           - verbose mode\n"
					"  -l c_file    - load calibration file for start\n"
					"  -L           - don't load calibration file for start\n"
					"  -D           - don't start ftdaemon\n"
					"Commands:\n"
					"  check                        - check TXT2 information and update conf. file\n"
					"  dump [n/all]                 - dump TXT2 information\n"
					"  fdterm on/off                - set default FDCAN terminator state\n"
			        "  bitrate speed                - set FDCAN bitrate\n"
					"  kill [signal]                - kill usual M4 clients\n"
					"  list                         - list usual M4 clients\n"
					"  name <name>                  - set TXT2 name for discovery\n"
					"  role single/master/slave     - set local controller role\n"
					"  stop                         - stop M4\n"
					"  start [-d]                   - start M4, -d for debug support\n"
					"  stat                         - statistic info\n"
					"  status                       - M4 status\n"
					"  power off                    - M4 power off\n"
					"  climit value                 - set 9V limit, 550 ma is default\n"
					"  reset                        - M4 reset\n"
					"  uid uid_value                - M4 uid\n"
			);
    exit(1);
}

static int IsRunning()
{
	int result = 0;
	char state[20];
	FILE *tf = fopen(stateFile, "r");

	if (tf == NULL)
		return result;

	fscanf(tf,"%16s", state);
	if (!strcmp(state, "running"))
		result = 1;

	fclose(tf);
    return result;
}

static void ProcessValue(char *name, char *value)
{
	if (verbose)
		printf("Process value %s = %s\n", name, value);

	if (!strcmp(name, "executable")) {
		if (executable == NULL)
			executable = strdup(value);
	} else if (!strcmp(name, "fdterm")) {
		if (!strcmp(value, "off")) {
			fdterm = 0;
		} else {
			fdterm = 1;
		}
	} else if (!strcmp(name, "role")) {
		if (!strncmp(value, "single", 6)) {
			role = 0;
		} else if (!strncmp(value, "master", 6)) {
			role = 1;
		} else if (!strncmp(value, "slave", 5)) {
			role = 2;
		}
	} else if (!strcmp(name, "bitrate")) {
		sscanf(value, "%ud", &bitrate);
	} else if (!strcmp(name, "buid")) {
		int v = 0;
		sscanf(value, "%i", &v);
		buid = v;
	} else if (!strcmp(name, "climit")) {
		sscanf(value, "%i", &climit);
	} else if (!strcmp(name, "name")) {
		strncpy(tname, value, sizeof(tname));
		tname[sizeof(tname) - 1] = 0;
	} else {
		fprintf(stderr, "Unknown value name : %s\n", name);
	}
}

static void ProcessLine(char *line)
{
    char *p, *n, *v;
    int c;

    for(p = line; (c = *p); p++) {
    	if (c != ' ' && c != '\t' && c != '\n')
    		break;
    }
    if (!c || !isalpha(c)) {
    	fprintf(stderr, "Wrong line : %s\n", line);
    	return;
    }

    /* value name is found */
    n = p;
    for(; (c = *p); p++) {
       	if (!isalnum(c))
       		break;
    }

    /* end name */
    *p++ = 0;
    if (!c) {
    	fprintf(stderr, "Wrong line : %s\n", line);
    	return;
    }
    if (c != '=') for(; (c = *p); p++) {
    	if (c != ' ' && c != '\t' && c != '\n')
    		break;
    }
    if (c != '='){
    	fprintf(stderr, "Wrong line : %s\n", line);
    	return;
    }

    for(; (c = *p); p++) {
        if (c != ' ' && c != '\t' && c != '\n')
        	break;
    }
    v = p;
    if (!c) {
    	fprintf(stderr, "Wrong line : %s\n", line);
    	return;
    }
    for(; (c = *p); p++) {
        if (c == ' ' || c == '\t' || c == '\n')
        	break;
    }
    *p = 0;
    ProcessValue(n, v);
}

static char *GetConfigurationFile()
{
	if (conffile == NULL) {
		if (access(HW_CONFIG, R_OK | W_OK) >= 0) {
			cname = HW_CONFIG;
			return cname;
		}
		char *home = getenv("HOME");
		if (home == NULL)
			home = ".";
		int dlen = strlen(home) + strlen(FT_CONF_FILE) + 1;
		cname = malloc(dlen);
		strcat(strcpy(cname, home), FT_CONF_FILE);
		if (access(cname, R_OK | W_OK) >= 0)
			return cname;

		cname = HW_CONFIG;
	} else {
		cname = conffile;
	}

	return cname;
}

static void ReadConfig()
{
	char *line = NULL;
	size_t len = 0;
	ssize_t nread;
	FILE *tf = fopen(cname, "r");

	if (tf == NULL) {
		fprintf(stderr, "Can't read configuration file : %s\n", cname);
		return;
	}

	/* Process lines */
	while((nread = getline(&line, &len, tf)) != -1) {
        if (line[0] == '#')
        	continue;
        ProcessLine(line);
	}

	free(line);
	fclose(tf);
}

static void SaveConfig()
{
	FILE *tf;

	if (!save)
		return;

	tf = fopen(cname, "w");
	if (executable != NULL)
		fprintf(tf, "executable=%s\n", executable);
	fprintf(tf, "fdterm=%s\n", fdterm ? "on" : "off");
	fprintf(tf, "bitrate=%d\n", bitrate);
	fprintf(tf, "climit=%d\n", climit);
	fprintf(tf, "role=%s\n", roles[role]);
	if (buid != kuid)
		fprintf(tf, "buid=%d\n", buid);
	if (tname[0] != 0)
		fprintf(tf, "name=%s\n", tname);

	fclose(tf);
}

static void sendCommand(char *fname, char *command)
{
	FILE *tf = fopen(fname, "w");
	if (tf == NULL) {
		perror(fname);
		exit(1);
	}
	fprintf(tf, command);
	fclose(tf);
}

static void Stop()
{
	sendCommand(stateFile, "stop");
}

static void Firmware()
{
	sendCommand(firmwareFile, executable);
}

static void Start()
{
	sendCommand(stateFile, "start");
}

static void SendI2cCommand(int id, uint8_t *command, int len, char *error)
{
	for(int tries = 5; tries > 0; tries--) {
		if (write(id, command, len) == len)
			return;
		usleep(10000);
	}

	perror(error);
}

static void FixI2cPower()
{
	if (rev01 >= 3) /* rev3 does not support i2c control for +9V */
		return;

	int id = open("/dev/i2c-2", O_RDWR);
	if (id < 0) {
		perror("I2c file");
		return;
	}

	if (ioctl(id, I2C_SLAVE, 0x60) < 0) {
		perror("I2c file");
		close(id);
		return;
	}

	/* Send limit */
	static uint8_t ecommand[] = {0x0a, 0xff}; /* reset errors */

	SendI2cCommand(id, ecommand, sizeof(ecommand), "Can't reset PWR errors");

	if (verbose)
		fprintf(stderr, "Setup %d ma 9V limit\n", climit);

	uint8_t lcommand[] = {0x03, climit / 50}; /* Set current limit */

	SendI2cCommand(id, lcommand, sizeof(lcommand), "I2c limit setup error");

	close(id);
}

static void CreateEndpoints()
{
	struct rpmsg_endpoint_info info;
    int i, id;

	if (access("/dev/rpmsg0", R_OK | W_OK) >= 0) {
		if (verbose)
			fprintf(stderr, "Rpmsg files already created\n");
		return;
	}

    /* First start sometimes requires more than 8 seconds ! */
    for (i = 0; i < 12; i++) { /* 12 seconds timeout */
        id = open(RPMSG_DEV, O_RDWR | O_NONBLOCK);
        if (id >= 0)
        	break;

        sleep(1);
    }

    if (id < 0) {
    	perror(RPMSG_DEV);
    	exit(1);
    }

    strncpy(info.name, RPMSG_SERVICE_NAME, 31);
    info.name[31] = '\0';
    info.dst = 0;

    for (i = 0; i < MAX_ENDPOINTS; i++) {
        /* Create endpoint i + 2 with file /dev/rpmsgI */
    	info.src = i + 2;
    	if (ioctl(id, RPMSG_CREATE_EPT_IOCTL, &info) < 0) {
    		perror("create endpoint");
    		exit(1);
    	}
    }

    close(id);
}

#define CALIB_ERROR return	/*!< Standard error processing in ftcalibcommon.h for ftconfig */
#include "ftcalibcommon.h"

static void LoadCalibration()
{

	/* Check rpmsg module is loaded */
	if (access(calib_file, R_OK) < 0) {
		if (verbose)
			fprintf(stderr, "Calibration file %s is unavailable\n", calib_file);
		return;
	}

	/* Map M4 memory */
	uint32_t offset = 0;
	int fd = open("/dev/m4mem", O_RDWR | O_SYNC);
	if (fd < 0) {
		perror("m4mem");
		fd = open("/dev/mem", O_RDONLY | O_CLOEXEC);
		offset = M4_ADDR; /* Physical address */
		if (fd < 0) {
			perror("mem");
			exit(1);
		}
	}

	void *xaddr = mmap(NULL, M4_MEM_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED,
			fd, (off_t) offset);
	if (xaddr == MAP_FAILED) {
		perror("mmap");
		exit(1);
	}

	/* Find mark in memory */
	struct ftcalib *ftc = FindMark(xaddr, M4_MEM_SIZE, 1);
	if (verbose)
		fprintf(stderr, "mark %p\n", (void*)ftc);

	if (ftc->version != FTCALIB_VER) {
		fprintf(stderr, "Firmware calibration version %d is unsupported\n", ftc->version);
		return;
	}

	LoadCalibFile(ftc, calib_file);
}

static void FullStart(char *value)
{
	void *txt;

	if (value == NULL || strcmp(value, "-d")) {
		/* Stop device unconditionally */
		Stop();
		sleep(STIMEOUT); /* Stop requires 0.5 sec timeout */

		/* Setup firware */
		Firmware();
		sleep(STIMEOUT);

		/* Start firmware */
		Start();
	}

	/* Fix limit for power, available for rev < 3 */
	FixI2cPower();

	/* Create endpoints */
	CreateEndpoints();

	/* Setup parameters */
	txt = ftop_alloc_txt("auto", 65000);
	if (txt == NULL) {
		if (verbose)
			fprintf(stderr, "Can't connect to TXT2\n");
		exit(1);
	}
	ftop_set_uid(txt, buid);
	ftop_set_terminator(txt, fdterm);
	ftop_can_set_bitrate(txt, bitrate);
	if (tname[0] != 0)
		ftop_txt_name(txt, tname);
	ftop_txt_mode(txt, role);
	ftop_free_txt(txt);
	if (loadcalib) {
		LoadCalibration();
	}
}

static void Check()
{
	struct ftobject_txt2 *txt;
	struct ft_shm_userver *userver;
	struct ft_shm_txt *current;
	int i;

	if (!IsRunning()){
		fprintf(stderr, "TXT2 is not running\n");
		return;
	}

	txt = ftop_alloc_txt("auto", 65000);
	if (txt == NULL) {
		if (verbose)
			fprintf(stderr, "Can't connect to TXT2\n");
		exit(1);
	}
	while (!txt->uaddr) sleep(1);

	userver = (void *)txt->header;
	/* Output information */
	fprintf(stderr, "UIDS:\n");
	for (i = 0; i < FT_MAX_TXT; i++) {
		ftuid uid = txt->suids[i];
		if (uid == 0)
			continue;
		fprintf(stderr, "uid %d 0x%8.8x\n", i, uid);
	}

	void* txts = (char *)userver + ((uint32_t)userver->header.txts - txt->uaddr);
	uint32_t txtaddr = *(uint32_t *)txts; /* Internal address */

	current = (void *)((char *)userver + (txtaddr - txt->uaddr));
    role = current->common.mode;
    nslave = current->common.nslave;
	strncpy(tname, current->common.device_name, sizeof(tname));
	tname[sizeof(tname) - 1] = 0;
    fprintf(stderr, "Role is %s\n", roles[role]);
    fprintf(stderr, "Name is %s\n", tname);

	ftop_free_txt(txt);
}

static void Dump(char *name)
{
	struct ftobject_txt2 *txt;
	struct ft_shm_userver *userver;
	int n = -1;
	int i;
	int imin, imax;

	if (name == NULL || !strcmp(name, "all")) {
		;
	} else if (name[0] >= '0' && name [0] <= '9' && name[1] == 0) {
        n = name[0] - '0';
	} else {
		fprintf(stderr, "Illegal parameter : %s\n", name);
		exit(1);
	}

	/* Implement it */
	if (!IsRunning()){
		fprintf(stderr, "TXT2 is not running\n");
		return;
	}

	txt = ftop_alloc_txt("auto", 65000);
	if (txt == NULL) {
		if (verbose)
			fprintf(stderr, "Can't connect to TXT2\n");
		exit(1);
	}

	if (n != -1) {
		imin = n;
		imax = n;
	} else {
		imin = 0;
		imax = FT_MAX_TXT - 1;
	}

	while (!txt->uaddr) sleep(1);

	/* Check firmware name */
	char *exec = NULL;
	size_t nn = 0;
	FILE *tf = fopen(firmwareFile, "r");
	getline(&exec, &nn, tf);
	fclose(tf);

	fprintf(stderr, "Current executable : %s\n", exec);

	userver = (void *)txt->header;
	struct ft_shm_header *header = txt->header;
	struct ft_shm_stat *ftstat = (struct ft_shm_stat *) ((char *) header
				+ ((uint32_t) header->stat - txt->uaddr));

	/* Current mode */
	fprintf(stderr, "Current mode:\n");
	fprintf(stderr, "Bitrate : %d\n", ftstat->bitrate);
	fprintf(stderr, "FDTerm  : %d\n", ftstat->fdterm);
	/* Output information */
	fprintf(stderr, "UIDS:\n");
	for (i = imin; i <= imax; i++) {
		ftuid uid = txt->suids[i];
		if (uid == 0)
			continue;
		fprintf(stderr, "uid %d 0x%8.8x\n", i, uid);
	}

	/* Output TXT2 controllers information */
	for (i = imin; i <= imax; i++) {
		int j;
		void* txts = (char *)userver + ((uint32_t)userver->header.txts - txt->uaddr);
		uint32_t txtaddr = ((uint32_t *)txts)[i]; /* Internal address */
		struct ft_shm_txt *s = (void *) ((char *) userver + (txtaddr - txt->uaddr));
		if (!s->common.firmware)
			continue;

		fprintf(stderr, "TXT2 %d\n", i);
		fprintf(stderr, "version %d\n", s->common.version);
		union {
			uint32_t firmware;
			uint8_t  fver[4];
		} fw = {.firmware = s->common.firmware};
		fprintf(stderr, "firmware %d.%d.%d.%d\n", fw.fver[3], fw.fver[2], fw.fver[1], fw.fver[0]);
		fprintf(stderr, "mode %d\n", s->common.mode);
		fprintf(stderr, "name %s\n", s->common.device_name);
		fprintf(stderr, "sn   %s\n", s->common.sn);
		if (!verbose)
			continue;

		/* Output config */
		fprintf(stderr, "  Config:\n");
		for (j = 0; j < FT_MAX_INPUTS; j++) {
			uint8_t code =  s->config.inputs.inputs[j][0];
			const char *tname = ftmop_get_typename(s->config.inputs.inputs[j][1]);

			fprintf(stderr, "  I%d(%s) - ", j + 1, tname);
			if (code & ftinp_ultrasonic) {
				fprintf(stderr, "ultrasonic\n");
			} else if (code == 0){
				fprintf(stderr, "unconfigured\n");
			} else {
				fprintf(stderr, (code & ftinp_analog) ? "analog" : "digital");
				fprintf(stderr, (code & ftinp_passive) ? "|passive" : "|active");
				fprintf(stderr, (code & ftinp_r2) ? "|r2\n" : "\n");
			}
		}
	}

	ftop_free_txt(txt);
}

static void FDTerm(char *state)
{
	void *txt;

	if (state == NULL) {
		fprintf(stderr, "FDCAN terminator state is %s\n",
				fdterm ? "on" : "off");
	} else if (!strcmp(state, "on")) {
		fdterm = 1;
	} else if (!strcmp(state, "off")) {
		fdterm = 0;
	} else {
		fprintf(stderr, "Illegal parameter : %s\n", state);
		exit(1);
	}

	if (!IsRunning()){
		fprintf(stderr, "TXT2 is not running\n");
		return;
	}

	/* Try to fix if running */
	txt = ftop_alloc_txt("auto", 65000);
	if (txt == NULL) {
		if (verbose)
			fprintf(stderr, "Can't connect to TXT2\n");
		exit(1);
	}
	ftop_set_terminator(txt, fdterm);
	ftop_free_txt(txt);
}

static void Bitrate(char *value)
{
	if (value == NULL) {
		fprintf(stderr, "CAN bitrate is %d\n", bitrate);
		return;
	}

	sscanf(value, "%ui", &bitrate);

	if (!IsRunning()) {
		fprintf(stderr, "TXT2 is not running\n");
		return;
	}

	void *txt = ftop_alloc_txt("auto", 65000);
	if (txt == NULL) {
		if (verbose)
			fprintf(stderr, "Can't connect to TXT2\n");
		exit(1);
	}
	ftop_can_set_bitrate(txt, bitrate);
	ftop_free_txt(txt);
}

static void Uid(char *value)
{
	if (value == NULL) {
		fprintf(stderr, "Uid is 0x%x %d\n", buid, buid);
		return;
	}

	buid = strtod(value, NULL); /* Problem with %i and strtol */

	if (!IsRunning()) {
		fprintf(stderr, "TXT2 is not running\n");
		return;
	}

	void *txt = ftop_alloc_txt("auto", 65000);
	if (txt == NULL) {
		if (verbose)
			fprintf(stderr, "Can't connect to TXT2\n");
		exit(1);
	}
	ftop_set_uid(txt, buid);
	ftop_free_txt(txt);
}

static void ListKill(char *value)
{
	extern int ftop_open_lock(void);

	struct flock lock = {.l_type = F_WRLCK, .l_whence = SEEK_SET, .l_len = 1, .l_start = 0};
	int mode = 0;
	int sig = 0;
	char *line = NULL;
	size_t len = 0;
	size_t nread;
	FILE *tf;

	if (value == NULL) { /* kill sigterm */
		sig = SIGTERM;
	} else if (value[0] == 0) { /* list */
		mode = 1;
	} else {
		sscanf(value, "%d", &sig);
	}

	int lid = ftop_open_lock();
	if (lid < 0) {
		perror("open lock file");
		exit(1);
	}

	for (int i = 0; i < FT_MAX_CLIENTS; i++) {
		lock.l_start = i;
		lock.l_pid = -1;
		lock.l_type = F_WRLCK;
		fcntl(lid, F_GETLK, &lock);
		if (lock.l_pid < 0)
			continue;
		if (mode || verbose) {
			fprintf(stderr, "client %d pid %d", i, lock.l_pid);
			char cmdline[6 + 5 + 8 + 3];
			sprintf(cmdline, "/proc/%d/cmdline", lock.l_pid);
			tf = fopen(cmdline, "r");
			if (tf != NULL) {
				if ((nread = getline(&line, &len, tf)) > 0) {
					for (size_t j = 0; j <= nread - 1; j++) if (line[j] == 0) line[j] = ' ';
					fprintf(stderr, " %s", line);
				}
				fclose(tf);
			}
			fprintf(stderr, "\n");
		}
		if (!mode)
			kill(lock.l_pid, sig);
	}

	close(lid);
}

static void Role(char *state, char *value)
{
	void *txt;

	(void)value;

	if (state == NULL) {
		fprintf(stderr, "Role %s", roles[role]);
	} else if (!strcmp(state, "single")) {
		role = 0;
	} else if (!strcmp(state, "master")) {
		role = 1;
	} else if (!strcmp(state, "slave")) {
		role = 2;
	} else {
		fprintf(stderr, "Illegal parameter : %s\n", state);
		exit(1);
	}

	if (state == NULL || !IsRunning()) /* TODO */
		return;

	txt = ftop_alloc_txt("auto", 65000);
	if (txt == NULL) {
		if (verbose)
			fprintf(stderr, "Can't connect to TXT2\n");
		exit(1);
	}
	ftop_txt_mode(txt, role);
	ftop_free_txt(txt);
}

static void Name(char *value)
{
	void *txt;

	if (value == NULL || value[0] == 0) {
		fprintf(stderr, "Name is %s\n", tname);
		return;
	}

	strncpy(tname, value, sizeof(tname));
	tname[sizeof(tname) - 1] = 0;

	if (!IsRunning()) {
		fprintf(stderr, "TXT2 is not running\n");
		return;
	}

	txt = ftop_alloc_txt("auto", 65000);
	if (txt == NULL) {
		if (verbose)
			fprintf(stderr, "Can't connect to TXT2\n");
		exit(1);
	}
	ftop_txt_name(txt, tname);
	ftop_free_txt(txt);
}

static int Stat()
{
	struct ftobject_txt2 *txt= ftop_alloc_txt("auto", 65000);
	if (txt == NULL) {
		if (verbose)
			fprintf(stderr, "Can't connect to TXT2\n");
		return 1;
	}

	/* Get statistic memory */
	struct ft_shm_header *header = txt->header;
	struct ft_shm_stat *ftstat = (struct ft_shm_stat *) ((char *) header
			+ ((uint32_t) header->stat - txt->uaddr));

	uint32_t runtime = ftstat->runtime;
	fprintf(stderr, "runtime              : %d ms\n", runtime);
	fprintf(stderr, "clients              : 0x%x\n", ftstat->clients);
	fprintf(stderr, "commands             : %d\n", ftstat->commands);
	fprintf(stderr, "ADC interrupts       : %d\n", ftstat->adccounter);
	fprintf(stderr, "SPI interrupts       : %d\n", ftstat->spi3counter);
	fprintf(stderr, "ULT interrupts       : %d\n", ftstat->usart6counter);
	fprintf(stderr, "ADI interrupts       : %d\n", ftstat->tim7counter);
	fprintf(stderr, "IPCC RX interrupts   : %d\n", ftstat->ipccrxcounter);
	fprintf(stderr, "IPCC TX interrupts   : %d\n", ftstat->ipcctxcounter);
	fprintf(stderr, "max motor scheduling : %d ms\n", ftstat->maxmotortasktime);
	fprintf(stderr, "max motor cycle time : %d ms\n", ftstat->maxmotortaskcycle);
	fprintf(stderr, "max hwloop time      : %d mks\n", ftstat->maxhwlooptime * 10);
	fprintf(stderr, "transmit packets     : %d\n", ftstat->cantx);
	fprintf(stderr, "receive packets      : %d\n", ftstat->canrx);
	fprintf(stderr, "Tx drop              : %d\n", ftstat->tdrop);
	fprintf(stderr, "Power +9V alerts     : %d\n", ftstat->pwralert);
	fprintf(stderr, "Motor driver alerts  : %d\n", ftstat->drvalert);
	fprintf(stderr, "Power +5V alerts     : %d\n", ftstat->servoalert);
	fprintf(stderr, "M0 start time        : %d\n", ftstat->m0maxtime);
	fprintf(stderr, "Bitrate              : %d\n", ftstat->bitrate);
	fprintf(stderr, "FDTerm               : %d\n", ftstat->fdterm);

	if (verbose) {
		/* Tasks info */
		uint32_t *task_times = (uint32_t *) ((char *) header
				+ ((uint32_t) header->task_times - txt->uaddr));

		printf("Tasks:\n");
		printf("motor task   : %10d %5.2f %%\n", task_times['M'], (double)task_times['M']/runtime);
		printf("ftserver     : %10d %5.2f %%\n", task_times['f'], (double)task_times['f']/runtime);
		printf("can engine   : %10d %5.2f %%\n", task_times['R'], (double)task_times['R']/runtime);
		printf("notification : %10d %5.2f %%\n", task_times['N'], (double)task_times['N']/runtime);
		printf("hwloop       : %10d %5.2f %%\n", task_times['h'], (double)task_times['h']/runtime);

	}

	ftop_free_txt(txt);
	return 0;
}

static int Status()
{
	char rpmsg_ep[] = "/dev/rpmsgX\0";
	int i;

	/* Check config file values */
	if (executable == NULL) {
		if (verbose)
			fprintf(stderr, "Executable is undefined\n");
		return 1;
	}

	/* Check rpmsg module is loaded */
	if (access("/dev/rpmsg_ctrl0", R_OK | W_OK) < 0) {
		if (verbose)
			fprintf(stderr, "Rpmsg raw module is unavailable\n");
		return 1;
	}

	/* Check endpoints */
	for (i = 0; i < MAX_ENDPOINTS; i++) {
		rpmsg_ep[sizeof(rpmsg_ep) - 3] = '0' + i;
		if (access(rpmsg_ep, R_OK | W_OK) < 0) {
			if (verbose)
				fprintf(stderr, "Rpmsg endpoint %s is unavailable\n", rpmsg_ep);
			return 1;
		}
	}

	/* Check firmware name */
	char *exec = NULL;
	size_t nn = 0;
	FILE *tf = fopen(firmwareFile, "r");
	getline(&exec, &nn, tf);
	fclose(tf);

	/* Remove eol */
	char *eol = index(exec, '\n');
	if (eol)
		*eol = 0;

	if (strcmp(exec, executable)) {
		if (verbose)
			fprintf(stderr, "Incorrect executable %s, must be %s\n", exec, executable);
		return 1;
	}

	/* Try to connect */
	struct ftobject_txt2 *txt = ftop_alloc_txt("auto", 65000);
	if (txt == NULL) {
		if (verbose)
			fprintf(stderr, "Can't connect to TXT2\n");
		return 1;
	}

	/* Check input activity */
	for (i = 0; i < 10; i++) {
		if (txt->txt != NULL)
			break;
		sleep(1);
	}
	if (txt->txt == NULL) {
		if (verbose)
			fprintf(stderr, "TXT2 hangs\n");
		sleep(1); /* Give a chance to output */
		kill(getpid(), 1);
		return 1;
	}

	ftop_free_txt(txt);
	if (verbose)
		fprintf(stderr, "TXT2 is OK\n");

	return 0;
}

static int Power(char *value)
{
	if (strcmp(value, "off")) {
		fprintf(stderr, "incorrect argument\n");
		return 1;
	}

	struct ftobject_txt2 *txt = ftop_alloc_txt("auto", 65000);
	if (txt == NULL) {
		if (verbose)
			fprintf(stderr, "Can't connect to TXT2\n");
		return 1;
	}
	ftop_power_down(txt);
	sleep(2);
	Stop();
	return 0;
}

static void Reset()
{
	struct ftobject_txt2 *txt = ftop_alloc_txt("auto", 65000);
	if (txt == NULL) {
		if (verbose)
			fprintf(stderr, "Can't connect to TXT2\n");
		return;
	}
	ftop_reset(txt);
}

/**
 * @brief Standard main procedure
 * @param [in] argc number of arguments
 * @param [in] argv arguments
 * @return Error code
 */
int main(int argc, char **argv)
{
	int dstart = 0;
	if (argc < 2)
		Usage();

	argv++;

	/* Process arguments */
	for (; argv[0] != NULL && argv[0][0] == '-'; argv++) {
		switch (argv[0][1]) {
		case 'c':
			if ((conffile = argv[1]) == NULL)
				Usage();
			argv++;
			break;
		case 's':
			save = 1;
			break;
		case 'e':
			if ((executable = argv[1]) == NULL)
				Usage();
			argv++;
			break;
		case 'v':
			verbose = 1;
			break;
		case 'L':
			loadcalib = 0;
			break;
		case 'l':
			if ((calib_file = argv[1]) == NULL)
				Usage();
			argv++;
			break;
		case 'D':
			startDaemon = 0;
			break;
		default:
			fprintf(stderr, "Unknown option: %s\n", argv[0]);
			break;
		}
	}

	GetConfigurationFile();
	if (verbose)
		fprintf(stderr, "Configuration file : %s\n", cname);

	/* Check TXT2 board revision */
	int rfid = open("/sys/bus/nvmem/devices/stm32-romem0/nvmem", O_RDONLY);
	if (rfid < 0) {
		perror("nvmem open");
	} else {
		if (lseek(rfid, 60 * 4, SEEK_SET) == -1) {
			perror("nvmem lseek");
		} else {
			uint32_t nboard = 0;
			if (read(rfid, &nboard, sizeof(nboard)) == sizeof(nboard)) {
				if (verbose)
					fprintf(stderr, "Board number : %d\n", nboard);

				buid = kuid = nboard;
			}
		}
		if (lseek(rfid, 67 * 4, SEEK_SET) == -1) {
			perror("nvmem lseek");
		} else {
			uint32_t rev = 0;
			if (read(rfid, &rev, sizeof(rev)) == sizeof(rev)) {
				rev01 = rev;
			}
		}
		close(rfid);
	}
	if (verbose) {
		fprintf(stderr, "Board revision : %d\n", rev01);
	}

	ReadConfig();

	if (executable == NULL)
		executable = rev01 ? "txt2-M4-rt-core-rev1.elf" : "txt2-M4-rt-core-rev0.elf";

	if (verbose)
		fprintf(stderr, "Executable file : %s\n", executable);

	if (argv[0] == NULL) {
		Usage();
	} else if (!strcmp(argv[0], "bitrate")) {
		Bitrate(argv[1]);
	} else if (!strcmp(argv[0], "check")) {
		Check();
	} else if (!strcmp(argv[0], "climit")) {
		if (argv[1] != NULL)
			sscanf(argv[1], "%i", &climit);
		FixI2cPower();
	} else if (!strcmp(argv[0], "dump")) {
		Dump(argv[1]);
	} else if (!strcmp(argv[0], "fdterm")) {
		FDTerm(argv[1]);
	} else if (!strcmp(argv[0], "kill")) {
		ListKill(argv[1]);
	} else if (!strcmp(argv[0], "list")) {
		ListKill("");
	} else if (!strcmp(argv[0], "name")) {
		Name(argv[1]);
	} else if (!strcmp(argv[0], "power")) {
		exit(Power(argv[1]));
	} else if (!strcmp(argv[0], "reset")) {
		Reset();
	} else if (!strcmp(argv[0], "role")) {
		Role(argv[1], argv[2]);
	} else if (!strcmp(argv[0], "stop")) {
		Stop();
	} else if (!strcmp(argv[0], "start")) {
		FullStart(argv[1]);
		dstart = startDaemon;
	} else if (!strcmp(argv[0], "status")) {
		exit(Status());
	} else if (!strcmp(argv[0], "stat")) {
		exit(Stat());
	} else if (!strcmp(argv[0], "uid")) {
		Uid(argv[1]);
	} else {
		Usage();
	}
	SaveConfig();
	if (dstart) {
		if (verbose)
			fprintf(stderr, "Start ftdaemon\n");
		execlp("ftdaemon", "ftdaemon", NULL);
		perror("execlp");
		exit(1);
	}
	return 0;
}

