/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file ftcalibcommon.h
 * @brief Common algorithms for calibration values update
 */

#ifndef FTINC_FTCALIBCOMMON_H_
#define FTINC_FTCALIBCOMMON_H_

#include <stddef.h>

static struct ftcalib *FindMark(void *addr, uint32_t size, int mode)
{
    char *maddr = addr;
    char *eaddr = maddr + size - sizeof(FTCALIB);

	for (; maddr < eaddr; maddr++) {
		if (!strcmp(maddr, FTCALIB)) {
			struct ftcalib *ftc = (void*) maddr;
			if (mode && !ftc->mmark)
				continue;

			return (struct ftcalib *) maddr;
		}
	}

    fprintf(stderr, "Can't find calibration mark\n");
    exit(2);
}

static void Copy1(struct ftcalib1 *in, struct ftcalib *out)
{
	out->adcs = in->adcs;
	out->adct1 = in->adct1;
	out->adct2 = in->adct2;

	for (int i = 0; i < 8; i++) {
		out->channel[i].div1 = in->adcd1;
		out->channel[i].div2 = in->adcd2;
		out->channel[i].val1 = in->adcr1;
		out->channel[i].val2 = in->adcr2;
	}
}

static void Copy2(struct ftcalib2 *in, struct ftcalib *out)
{
	int n = (sizeof(*in) - offsetof(struct ftcalib, version)) / sizeof(uint32_t);
	uint32_t *from = &in->version;
	uint32_t *to   = &out->version;

	from++; to++;

	for (int i = 1; i < n; i++) *to++ = *from++;
}

static void Copy3(struct ftcalib3 *in, struct ftcalib *out)
{
	int n = (sizeof(*in) - offsetof(struct ftcalib, version)) / sizeof(uint32_t);
	uint32_t *from = &in->version;
	uint32_t *to   = &out->version;

	from++; to++;

	for (int i = 1; i < n; i++) *to++ = *from++;
}

static void Copy(struct ftcalib *in, struct ftcalib *out)
{
	int n = (sizeof(*out) - offsetof(struct ftcalib, version) ) / sizeof(uint32_t);
	uint32_t *from = &in->version;
	uint32_t *to   = &out->version;

	from++; to++;

	for (int i = 1; i < n; i++) *to++ = *from++;
}

/* Load calibration parameters from file */
static void LoadCalibFile(struct ftcalib *ftc, char *fname)
{
	struct ftcalib in = {0};

	if (ftc->version != FTCALIB_VER) {
		fprintf(stderr, "Firmware calibration version %d is unsupported\n", ftc->version);
		CALIB_ERROR;
	}

	FILE *tf = fopen(fname, "r");
	if (tf == NULL) {
		perror("fopen");
		CALIB_ERROR;
	}

	size_t clen = fread(&in, 1, sizeof(in), tf);
	fclose(tf);

	if (strcmp(in.ftmark, FTCALIB)) {
		fprintf(stderr, "Incorrect mark in the %s file, can't load\n", fname);
		CALIB_ERROR;
	}

	if (ftc->uid != in.uid) {
		fprintf(stderr, "Incorrect uid 0x%x in %s file, can't load\n", in.uid, fname);
		CALIB_ERROR;
	}

	switch(in.version) {
	case 1:
		if (clen != sizeof(struct ftcalib1)) {
			fprintf(stderr, "Incorrect %s file length, can't load\n", fname);
			CALIB_ERROR;
		}
		Copy1((struct ftcalib1 *)&in, ftc);
		break;
	case 2:
		if (clen != sizeof(struct ftcalib2)) {
			fprintf(stderr, "Incorrect %s file length, can't load\n", fname);
			CALIB_ERROR;
		}
		Copy2((struct ftcalib2 *)&in, ftc);
		break;
	case 3:
		if (clen != sizeof(struct ftcalib3)) {
			fprintf(stderr, "Incorrect %s file length, can't load\n", fname);
			CALIB_ERROR;
		}
		Copy3((struct ftcalib3 *)&in, ftc);
		break;
	case 4:
		if (clen != sizeof(struct ftcalib)) {
			fprintf(stderr, "Incorrect %s file length, can't load\n", fname);
			CALIB_ERROR;
		}
		Copy(&in, ftc);
		break;
	default:
		fprintf(stderr, "Incorrect ftcalib version in %s file, can't load\n", fname);
		CALIB_ERROR;
	}

}

#endif /* FTINC_FTCALIBCOMMON_H_ */
