/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file ftlibtypes.h
 * @brief Linux implementation types
 */

#ifndef FTINC_FTLIBTYPES_H_
#define FTINC_FTLIBTYPES_H_

#include "ftshmem.h"

#include <pthread.h>

/** Magic values */
enum ftobjects_magics {
    ft_txt2_magic = 0x0f06868c,               /*!< ftobject txt2 magic value */
    ft_resistor_magic = 0x20adc185,           /*!< ftobject resistor magic value */
    ft_termistor_magic = 0x71fc1ab0,          /*!< ftobject termistor magic value */
    ft_switch_magic = 0x685f67a8,             /*!< ftobject switch magic value */
    ft_voltmeter_magic = 0x717730b3,          /*!< ftobject voltmeter magic value */
    ft_color_sensor_magic = 0x231dba98,       /*!< ftobject color_sensor magic value */
    ft_trail_follower_magic = 0x53fae271,     /*!< ftobject trail_follower magic value */
    ft_lamp_magic = 0x5a2edf91,               /*!< ftobject lamp magic value */
    ft_motor_magic = 0x08ee2ea2,              /*!< ftobject motor magic value */
    ft_motor_step_counter_magic = 0x53f1affa, /*!< ftobject motor step counter magic value */
    ft_counter_magic = 0x53c93e1c,            /*!< ftobject counter magic value */
    ft_ultrasonic_magic = 0x75925f4e,         /*!< ftobject ultrasonic magic value */
	ft_servo_magic = 0x6d033c22,              /*!< ftobject servo magic value */
	ft_digital_magic = 0x59864c01,            /*!< ftobject digital magic value */
	ft_encoder_magic = 0x5d17ca02,            /*!< ftobject encoder magic value */
};

/*!
 * Common connection object
 */
typedef struct {
   ftobject_magic magic; /*!< ftobject magic value */
   uint8_t conn;		 /*!< ftobject connection */
} ftobject_connection;

/*!
 * Universal inputs I1-I8
 */
typedef union {
    ftobject object;				/*!< ftobject */
    ftobject_connection connection;	/*!< ftobject connection */
} ftobjInput;

/*!
 * Universal motor
 */
typedef struct {
    ftobject_magic magic;  /*!< ftobject magic value */
    uint8_t conn;          /*!< ftobject connection */
    uint8_t counter;       /*!< encoder for motor */
    int16_t speed;         /*!< signed speed */
    uint16_t cmdid;        /*!< cmdid */
} ftobject_motor;

/*!
 * Servo motor
 */
typedef struct {
    ftobject_magic magic;	/*!< ftobject magic value */
    uint8_t conn;			/*!< ftobject connection */
    uint8_t pad;			/*!< padding */
    uint16_t speed;         /*!< unsigned speed */
} ftobject_smotor;

/*!
 * Universal outputs O1-O8
 */
typedef union {
    ftobject object;                /*!< any ftobject */
    ftobject_connection connection; /*!< any connction */
    ftobject_motor motor;           /*!< any motor */
} ftobjOutput;

/*!
 * Universal counters C1-C4
 */
typedef union {
    ftobject object;                /*!< any ftobject */
    ftobject_connection connection; /*!< any connction */
} ftobjCounter;

/*!
 * Servo motors S1-S4
 */
typedef union {
    ftobject object;                /*!< any ftobject */
    ftobject_connection connection; /*!< any connction */
    ftobject_smotor smotor;         /*!< servo motor */
} ftobjServo;

struct ftobject_txt2; // Forward

/**
 * Slave access structure
 */
struct ftobject_slave {
    struct ftobject_txt2 *txt2; /*!< TXT2 slave */
    ftuid uid;                  /*!< TXT2 slave UID */
    unsigned char caddr;        /*!< CAN addr */
};

/*!
 * TXT2 controller events implementation
 */
struct fttxt_events {
    uint8_t counts[ft_event_max]; /*!< conters for different event types */
};

/*!
 * TXT2 controller dynamic settings for servos implementation
 */
typedef struct ftobject_txt2_dynamic_servo_settings {
    uint16_t min;
    uint16_t max;
    uint8_t updated;
} ftobject_txt2_dynamic_servo_settings;

/*!
 * TXT2 controller dynamic settings implementation
 */
typedef struct ftobject_txt2_dynamic_settings {
    ftobject_txt2_dynamic_servo_settings servo[FT_MAX_SERVO];
} ftobject_txt2_dynamic_settings;

/*!
 * TXT2 controller implementation
 */
typedef struct ftobject_txt2 {
    ftobject_magic magic;  /*!< ftobject magic value */
    struct ftop_transport *transport; /*!< transport */
    struct ftobject_txt2 *main; /*!< pointer to the main controller */
    pthread_t thread;           /*!< communication thread */
    pthread_t sync_settings_thread;  /*!< settings thread */
    int cmdid;                  /*!< next cmdid */
    int memid;                  /*!< memory file id */

    /* Common TXT2 info */
    int version;                /*!< version */
    int firmware;               /*!< firmware */
    ftuid uid;                  /*!< processor uid */
    int mode;                   /*!< TXT2 controller mode */
    unsigned int nslaves;       /*!< number of slaves */
    char name[FT_NAME_SIZE];    /*!< TXT2 name */

    /* TXT2 connection */
    int canid;                   /*!< CAN id */

    /* Last fault id */
    int faultid;				/*!< fault id to check M4 alarm events */

    struct ftobject_slave slaves[FT_MAX_SLAVES]; /*!< Slaves info */

    /* Current structure for TXT2 objects */
    ftobjInput inputs[FT_MAX_INPUTS];       /*!< I1-I8 objects */
    ftobjOutput outputs[FT_MAX_OUTPUTS];    /*!< O1-O8 objects */
    ftobjCounter counters[FT_MAX_COUNTERS]; /*!< C1-C4 objects */
    ftobjServo servo[FT_MAX_SERVO];         /*!< S1-S3 objects */

    /* Event processing */
    pthread_mutex_t emutex;					/*!< mutex for M4 events */
    pthread_cond_t  econd;					/*!< conditional variable for event waiting */

    struct fttxt_events events;             /*!< events counters */

    /* New config for us */
    int connusage;                          /*!< used connections */
    struct ft_config new_config;            /*!< new config for M4 */
    int configid;                           /*!< next config id */

    uint32_t uaddr;                          /*!< uids addr dor shared memeory access */
    struct ft_shm_header *header;            /*!< shared memory header */
    ftvalues *cvalues;                       /*!< current values shared memory access */
    ftuid *suids;                            /*!< user level shared memory uids pointer */
    struct ft_shm_txt *txt;                  /*!< shared memory TXT2 structure */

    /* Current values from M4 */
    ftvalues current_values;                 /*!< Current values copy */

    ftobject_txt2_dynamic_settings dynamic_settings;

} ftobject_txt2;

#endif /* FTINC_FTLIBTYPES_H_ */
