/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @addtogroup Common_FT_API_Implementation_Includes
 * @{
 */

/**
 * @file ftserver.h
 * @brief M4 server specific procedures
 */
#ifndef _FTOP_SERVER
#define _FTOP_SERVER

#include "fttransport.h"
#include "ftintshmem.h"

/**
 * @addtogroup Host_Communication
 * @{
 */

/* Server specific */

/**
 * @brief Main procedure for command server task to process commands from A7 clients
 */
int ftop_start_server();

/**
 * Client specific motor information
 */
typedef struct ftop_motor_for_client {
	uint32_t mcmdid; /*!< motor cmdid for comparision */
	uint16_t lcmdid; /*!< Last executed client motor cmdid */
	uint16_t ncmdid; /*!< New stared client motor cmdid */
} ftop_motor_for_client;

/**
 * Client structure
 */
typedef struct ftop_client {
	int id;                     /*!< client id */
	int socket;                 /*!< client "socket" */
	uint32_t configid;          /*!< last configid */
	uint32_t ltick;				/*!< last event tick */
	uint8_t opened[FT_MAX_TXT]; /*!< ping processing */
	uint32_t usage[FT_MAX_TXT]; /*!< usage for hardware resources */
	struct ftop_motor_for_client motor_cmdid[FT_MAX_TXT][FT_MAX_MOTORS]; /*!< client specific motor information */
} ftop_client;

/**
 * @brief Process new command for remote TXT2 controller
 * @param ftc A7 client
 * @param buffer buffer with command
 */
void ftProcessRequest(ftop_client *ftc, unsigned char *buffer);

/**
 * @brief Get shared memory structure for the slave controller
 * @param num controller number
 * @return shared memory structure pointer for the slave controller
 */
struct ft_shm_txt *getTXT(int num);

/**
 * @brief Process control command from master
 * @param stxt remote master structure
 * @param op new operation
 */
void ProcessRemoteOp(struct ft_master_remote *stxt, struct ft_remote_op *op);

/**
 * @brief Prepare TXT2 controller values for master
 * @param values buffer for values
 */
void getValues(struct ft_remote_values *values);

/**
 * @brief Process new config from local/remote source
 * @param cusage resources usage value for config
 * @param c new config
 */
void ProcessNewConfig(uint32_t cusage, struct ft_config *c);

/**
 * @brief Get copy of operation for remote slave, called by remote engine
 * @param stxt external TXT2 controller
 * @param op buffer for copy
 */
void getOp(struct ft_shm_txt_remote *stxt, struct ft_remote_op *op);

/**
 * @brief Check uid/suid/cid combination
 * @param uid master controller uid
 * @param suid slave controller uid
 * @param cid CAN specific number from slave controller
 * @return shared memory structure pointer for the slave controller
 */
struct ft_shm_txt *Ext_TXT2_Check(ftuid uid, ftuid suid, int cid);

/**
 * @brief Initialization procedure
 */
void ftInitServer();

/**
 * @brief Procedure to calculate resources usage for all clients and master
 * @param ntxt TXT2 controller number, 0 - local
 * @return Resources usage for selected TXT2 controller
 **/
uint32_t CalculateTotalHWUsage(int ntxt);

/** @} */

#endif /* _FTOP_SERVER */

/** @} */
