/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file ftmotortask.c
 * @brief Motor Task code
 */

#include "ftconfig.h"
#include "ftservermem.h"
#include "ftosdep.h"
#include "main.h"
#include "DRV8908.h"

#include <openamp_log.h>
#include <string.h>
#include <FreeRTOS.h>
#include <cmsis_os.h>
#include <semphr.h>

#define MC_BREAK_START  45	/*!< initial value to stat motor break */
#define MC_BREAK_RANGE  25  /*!< maximum value to stop motor */
#define MC_CHECK_TIME 1     /*!< switch on motor task check time */

#define USE_MCOLLECTOR 0	/*!< switch on motor collector debug */

#if USE_MCOLLECTOR
#define MAX_COLLECTOR 1024
static struct {
	uint16_t tstamp;
	int      remainnorm;
	uint16_t sc;
	uint16_t remain;
	uint16_t delta;
	uint16_t d0, d1;
} Collector[MAX_COLLECTOR];
static int iCollector;
#endif

extern OS_MUTEX(out_mutex);
extern OS_MUTEX(drv8908_mutex);

/**
 * @brief High priority task to control motors
 * @param info unused
 */
void *MotorTaskStart(void *info)
{
	int i;
	uint16_t duty[FT_MAX_OUTPUTS];
#if MC_CHECK_TIME
	uint32_t mx = 0;
	uint32_t mt = 0;
	uint32_t ft = 0;
	uint32_t st = 0;
	uint32_t ut = 0;
	int32_t mtime = HAL_GetTick();
#endif
	uint32_t tprev = HAL_GetTick();

	while(1) {
		uint32_t sum[3] = {0};
		uint32_t pos[FT_MAX_MOTORS];
		uint8_t reached[FT_MAX_MOTORS] = {0};
		uint8_t group[FT_MAX_MOTORS] = {0};
		uint8_t m[3] = {0};
		uint8_t r[3] = {0};
		uint8_t mask = 0;

		OS_LOCK(out_mutex);
		uint32_t ntime = HAL_GetTick();
		if ((ntime - tprev) > 2) {
			log_info("motor task hang\n");
		}
		tprev = ntime;
#if MC_CHECK_TIME
		uint32_t mm = ntime - mtime;
		if (mm > mx) {
			txt2_server.stat.maxmotortasktime = mx = mm;
		}
#endif

		/* Copy user selected PWM duty to temporary variables */
		memcpy(duty, txt2_server.current.opwm, sizeof(duty));

		/* Distance break regulation */
		for (i = 0; i < FT_MAX_COUNTERS; i++) {
			uint8_t ig = 2;

			/* Add motor to group */
			if (txt2_server.mmask & (1 << i)) {
				ig = 0;
			} else if (txt2_server.mmask & (0x10 << i)) {
				ig = 1;
			}
			group[i] = ig;
			m[ig]++;
			pos[i] = txt2_server.current.mcommand[i].distance;

			if (txt2_server.break_info[i].reached) {
				/* Distance is reached or overshot => Full break */
				reached[i] = 1;
				duty[2 * i] = 0;
				duty[2 * i + 1] = 0;

				/* collect information for synchronization */
				r[ig]++;
				sum[ig] += pos[i];

				continue;
			}

			/* Current distance */
			uint32_t sc = txt2_server.counters[i] - txt2_server.fcounters[i];

			/* Collect information for synchronization */
			pos[i] = sc;
			sum[ig] += sc;

			/* No distance => no break */
			if (txt2_server.current.mcommand[i].distance == 0)
				continue;

			int32_t remain = (int)txt2_server.current.mcommand[i].distance - sc;
			uint32_t delta;

			/* Use time between encoder pulses as inverted speed */
			if (sc != txt2_server.break_info[i].lcount) {
				/* New pulse */
				txt2_server.break_info[i].dtime = ntime - txt2_server.break_info[i].ltime;
				txt2_server.break_info[i].ltime = ntime;
				txt2_server.break_info[i].lcount = sc;
			} else {
				/* No pulse state, increase delta time */
				delta = ntime - txt2_server.break_info[i].ltime;
				if (delta > txt2_server.break_info[i].dtime) {
					txt2_server.break_info[i].dtime = delta;
				}
			}

			delta = txt2_server.break_info[i].dtime;

			/* Check if distance is reached */
			if (remain <= 0) {
				/* Distance is reached or overshot => Full break */
				duty[2 * i] = 0;
				duty[2 * i + 1] = 0;

				reached[i] = 1;
				txt2_server.break_info[i].reached = 1;
				r[ig]++;
				continue;
			}

			/* Normal delta is 0-22, fix very small/large values */
			if (delta < 3) {
				delta = 3;
			} else if (delta > 128) {
				delta = 128;
			}

			/*
			 * Calculate speed normalized remaning distance
			 * remainnorm is proportional to distance/speed
			 */
			uint32_t remainnorm = remain * delta;

			/* No reason for break mode ? */
			if (remainnorm > MC_BREAK_START && !txt2_server.break_info[i].breakmode) {
				continue;
			}

#if USE_MCOLLECTOR
			iCollector = (iCollector + 1) & (MAX_COLLECTOR - 1);
			Collector[iCollector].tstamp = ntime;
			Collector[iCollector].remainnorm = remainnorm;
			Collector[iCollector].sc = sc;
			Collector[iCollector].remain = remain;
			Collector[iCollector].delta = delta;
#endif /* USE_MCOLLECTOR */

			txt2_server.break_info[i].breakmode = 1;

			/* Inertia range */
			if (remainnorm <= MC_BREAK_RANGE) {
				duty[2 * i] = 0;
				duty[2 * i + 1] = 0;
				continue;
			}

			/* Recalculate speed */
			int speed = (remainnorm - MC_BREAK_RANGE) * 4;
			if (duty[2 * i] > speed) {
				duty[2 * i] = speed;
			}
			if (duty[2 * i + 1] > speed) {
				duty[2 * i + 1] = speed;
			}
			mask |= 1 << i;
		}

#if MC_CHECK_TIME
		uint32_t ftime = HAL_GetTick();
		mm = ftime - ntime;
		if (mm > ft) {
			ft = mm;
		}
#endif

		/*
		 * Synchronization part.
		 * Simple sync regulation without master selection
		 */
		for (i = 0; i < FT_MAX_COUNTERS; i++) {

			if (reached[i])
				continue;

			int ig = group[i];
			if (ig == 2) {
				/* Skip motors without group */
				continue;
			}

			/* Distance from medium point */
			int32_t sync_err = pos[i] * m[ig] - sum[ig];
			if (!sync_err)
				continue;

			uint32_t ifix;
			if (sync_err < 0) { /* Try to increase speed */
				sync_err = -sync_err;
				if (sync_err > 16)
					sync_err = 16;
				ifix = 64 + 4 * sync_err;
			} else { /* Try to decrease speed */
				if (sync_err > 16)
					sync_err = 16;
				ifix = 64 - 4 * sync_err;
			}

			/* Recalulate duty values */
			duty[2 * i]     = (duty[2 * i] * ifix) >> 6;
			duty[2 * i + 1] = (duty[2 * i + 1] * ifix) >> 6;
			mask  |= (1 << i);
		}

#if MC_CHECK_TIME
		uint32_t stime = HAL_GetTick();
		mm = stime - ftime;
		if (mm > st) {
			st = mm;
		}
#endif

		/* Copy modified duty values to DRV8908 */
		OS_LOCK(drv8908_mutex);
		DRV8908_set_pwms(duty, txt2_server.current.motormode);
		OS_UNLOCK(drv8908_mutex);

#if MC_CHECK_TIME
		uint32_t utime = HAL_GetTick();
		mm = utime - stime;
		if (mm > ut) {
			ut = mm;
		}
#endif

		/* Check reached */
		for (i = 0; i < FT_MAX_COUNTERS; i++) {
			if (!reached[i])
				continue;

			int ig = group[i];
			if (ig == 2 || m[ig] == r[ig]) {
				/*
				 * Set end markers for nongroup motors and "reached" groups
				 * RoboPro requirement: any motor on group can be marked as reached
				 * after all motors in group.
				 */
				txt2_server.current.mcommand[i].lcmdid =
						txt2_server.current.mcommand[i].ncmdid;
			}
		}

		OS_UNLOCK(out_mutex);

#if MC_CHECK_TIME
		uint32_t xtime = HAL_GetTick();
		mtime = xtime;
		xtime -= ntime;
		if (xtime > mt) {
			txt2_server.stat.maxmotortaskcycle = mt = xtime;
		}
#endif
		osDelay(1);
	}

    return NULL;
}


