/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @defgroup Firmware_Implementation_Includes
 * @{
 */

/**
 * @file DRV8908.h
 * @brief DRV8908 constants and procedures
 */

#ifndef FTSRC_DRV8908_H_
#define FTSRC_DRV8908_H_

#include "fttypes.h"

/* DRV8908-Q1 has different registers map in different revisions */

/**
 * @addtogroup Hardware_Includes
 * @{
 */

/* Typedef */
struct ft_faults; /*!< Forward declaration */

/** DRV8908 registers, new map */
enum DRV89XX_registers {
	DRV_IC_STAT,        /* RD */
	DRV_OCP_STAT_1,
	DRV_OCP_STAT_2,
	DRV_OCP_STAT_3,      /* Reserved */
	DRV_OLD_STAT_1,
	DRV_OLD_STAT_2,
	DRV_OLD_STAT_3,      /* Reserved */
	DRV_CONFIG_CTRL,    /* RW */
	DRV_OP_CTRL_1,
	DRV_OP_CTRL_2,
	DRV_OP_CTRL_3,      /* Reserved */
	DRV_PWM_CTRL_1,
	DRV_PWM_CTRL_2,
	DRV_FW_CTRL_1,
	DRV_FW_CTRL_2,      /* Reserved */
	DRV_PWM_MAP_CTRL_1,
	DRV_PWM_MAP_CTRL_2,
	DRV_PWM_MAP_CTRL_3, /* Reserved */
	DRV_PWM_MAP_CTRL_4, /* New release */
	DRV_PWM_FREQ_CTRL = DRV_PWM_MAP_CTRL_4, /* 0x12 */
	DRV_PWM_FREQ_CTRL_1,
	DRV_PWM_FREQ_CTRL_2,
	DRV_PWM_DUTY_CTRL_1,
	DRV_PWM_DUTY_CTRL_2,
	DRV_PWM_DUTY_CTRL_3,
	DRV_PWM_DUTY_CTRL_4,
	DRV_PWM_DUTY_CTRL_5,
	DRV_OLD_CTRL_1 = DRV_PWM_DUTY_CTRL_5,
	DRV_PWM_DUTY_CTRL_6,
	DRV_OLD_CTRL_2 = DRV_PWM_DUTY_CTRL_6,  /* 0x1a */
	DRV_PWM_DUTY_CTRL_7,
	DRV_PWM_DUTY_CTRL_8,
	DRV_SR_CTRL_1,
	DRV_SR_CTRL_2,      /* Reserved */
	DRV_OCP_OLD_CTRL_1,
	DRV_OCP_OLD_CTRL_2,
	DRV_OCP_OLD_CTRL_3,
	DRV_OCP_OLD_CTRL_4,
	DRV_OCP_OLD_CTRL_5,
	DRV_OCP_OLD_CTRL_6
};

/** DRV8908 IC_STAT bits */
enum DRV89XX_StatBits{
	DRV_STAT_NPOR = 1,
	DRV_STAT_OVP = (1 << 1),
	DRV_STAT_UVLO = (1 << 2),
	DRV_STAT_OCP = (1 << 3),
	DRV_STAT_OLD = (1 << 4),
	DRV_STAT_OTW = (1 << 5),
	DRV_STAT_OTSD = (1 << 6),
	DRV_STAT_MASK = 0x3f,
	DRV_STAT_SAVE = 0xc0,
	DRV_STAT_EVT = (DRV_STAT_OVP | DRV_STAT_UVLO | DRV_STAT_OCP | DRV_STAT_OLD
			| DRV_STAT_OTW | DRV_STAT_OTSD)
};

/* Exported functions */

/**
 * @brief Copy alert pins and DRV89xx registers to shared memory structure
 * @param f buffer for faults info
 */
void DRV8908_Check_Faults(struct ft_faults *f);

/**
 * @brief Power foen procedure for DRV8908 and MP28167GQ +9V25, switch off 9V25
 */
void DRV8908_Stop();

/**
 * @brief Set PWM procedure for any DRV89xx
 * @param n 0-7 for O1-O8
 * @param pwm new PWM value 0-255
 */
void DRV8908_set_pwm(unsigned int n, uint8_t pwm);

/**
 * @brief The procedure to setup motor speed for any DRV89xx
 * @param n motor number 0, 2, 4 or 6
 * @param npwm two PWM values 0-512 for first and second pins
 */
void DRV8908_motor_speed(unsigned int n, ft_pwm_type *npwm);

/**
 * @brief The procedure to set all PWMs for any DRV89xx
 * @param pwms new 8 PWM values 0-512
 * @param motormode motor coast mode
 */
void DRV8908_set_pwms(uint16_t *pwms, uint8_t motormode);

#endif /* FTSRC_DRV8908_H_ */

/** @} */

/** @} */

