/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @defgroup Firmware_Implementation_Includes
 * @{
 */

/**
 * @file HWFDCAN.h
 * @brief Common FDCAN hardware procedures
 */

#ifndef INC_HWFDCAN_H_
#define INC_HWFDCAN_H_

#include <stdint.h>

/* Defines */

/* Can addrs */
#define FT_CAN_BASE_ID           128							/*!< CAN base address for range */
#define FT_CAN_ID_RANGE          (FT_MAX_TXT > 8 ? 32 : 16)		/*!< CAN range, must be >= 2 * FT_MAX_TXT */
#define FT_CAN_DISCOVERY_ID      (FT_CAN_BASE_ID + 1)			/*!< CAN discovery address */
#define FT_CAN_MARKER_ID         (FT_CAN_BASE_ID + 31)			/*!< TTCAN marker address */
#define FT_CAN_SLAVE_TX_ADDR(n)  (FT_CAN_DISCOVERY_ID + 2 * (n)) /*!< CAN slave address */
#define FT_CAN_MASTER_TX_ADDR(n) (FT_CAN_BASE_ID + 2 * (n))		/*!< CAN slave address */

/* Slave blocks */
#define FT_CAN_TXN_MARKER	  (0)	/*!< FDCAN Tx buffer for master */
#define FT_CAN_TXN_DISCOVERY  (0)	/*!< FDCAN Tx buffer for discovery */
#define FT_CAN_TXN_SLAVE(n)   (n)	/*!< FDCAN Tx buffer for slave */
#define FT_CAN_TXN_VALUES(n)  (n)	/*!< FDCAN Tx buffer for values */
#define FT_CAN_TXN_INFO(n)    (n)	/*!< FDCAN Tx buffer for info */
#define FT_CAN_TXN_VALUES(n)  (n)	/*!< FDCAN Tx buffer for values */

/* Master blocks,  n = 1,9 */
#define FT_CAN_TXN_CALIBRATION 1			 /*!< FDCAN Tx buffer for calibration */
#define FT_CAN_TXN_SADDR(n)    (2 * (n))	 /*!< FDCAN Tx buffer for set address */
#define FT_CAN_TXN_CONFIG(n)   (2 * (n))	 /*!< FDCAN Tx buffer for config */
#define FT_CAN_TXN_OP(n)       (1 + 2 * (n)) /*!< FDCAN Tx buffer for operation */

/* Tx bit blocks */
#define FT_CAN_TX_MARKER	   (1 << FT_CAN_TXN_MARKER)		 /*!< bit mask for marker */
#define FT_CAN_TX_DISCOVERY    (1 << FT_CAN_TXN_DISCOVERY)   /*!< bit mask for discovery */
#define FT_CAN_TX_CALIBRATION  (1 << FT_CAN_TXN_CALIBRATION) /*!< bit mask for calibration */

/* n = 1,9 */
#define FT_CAN_TX_SADDR(n)     (1 << FT_CAN_TXN_SADDR(n))	 /*!< bit mask for saddr */
#define FT_CAN_TX_CONFIG(n)    (1 << FT_CAN_TXN_CONFIG(n))	 /*!< bit mask for config */
#define FT_CAN_TX_OP(n)		   (1 << FT_CAN_TXN_OP(n))		 /*!< bit mask for operation */
#define FT_CAN_TX_VALUES(n)    (1 << FT_CAN_TXN_VALUES(n))	 /*!< bit mask for values */

/**
 * @addtogroup Hardware_Includes
 * @{
 */

/* Exported functions */

/**
 * @brief Init FDCAN GPIO pins: standby and sleep
 */
void HWFDCAN_init();

/**
 * @brief Select FDCAN terminator mode
 * @param mode new terminator mode: 0 - no, 1 - use terminator
 */
void HWFDCAN_set_terminator(int mode);

/**
 * @brief Selected new FDCAN bitrate
 * @param bitrate new bitrate (hz), for example, 1000000
 */
void HWFDCAN_set_bitrate(int bitrate);

/**
 * @brief Selected new FDCAN mode
 * @param mode new mode: 0 - single, 1 - master, 2 - slave
 */
void HWFDCAN_set_mode(int mode);

/**
 * @brief Add new selected address to filter for slave
 * @param canAddr CAN id
 */
void HWFDCAN_AddAddressToFilter(int canAddr);

/**
 * @brief Send packet procedure
 * @param packet packet buffer
 * @param size buffer size
 * @param addr CAN id
 */
void HWFDCAN_send_packet(uint8_t *packet, int size, int addr);

/**
 * @brief Send packet procedure
 * @param packet packet buffer
 * @param size buffer size
 * @param addr CAN id
 * @param nbuffer buffer for output
 */
void HWFDCAN_set_buffer(uint8_t *packet, int size, int addr, int nbuffer);

#endif /* INC_HWFDCAN_H_ */

/** @} */

/** @} */
