/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file DRV8908.c
 * @brief Control DRV8908 PWM device
 */

#include "DRV8908.h"

#include "main.h"
#include "ftshmem.h"

#include "cmsis_os.h"
#include "openamp_log.h"

/* Defines */

#define USE_ICOLLECTOR 0 /*!< Use icollector for debug */
#define PACKET_LENGTH 4	 /*!< Currect RXP/TXP packet lenth (must be compatible with cubemx) */

/* External variables */

int firstMotorCheck = 0; /*!< 0 - wait nonzero, 1 - wait counter, 2 - wait zero */
uint32_t motorCheckTick; /*!< initial nonzero time for M1 PWM motor */
extern SPI_HandleTypeDef hspi3; /*!< SPI3 HAL handle */

/* Private variables */

static int initok = 0;
static uint32_t tstart;
static uint32_t tend;
static uint32_t tmax;
static uint32_t rdiff;

#if USE_ICOLLECTOR
#define MAX_ICOLLECTOR 256
static struct {
	uint32_t itime;
	uint32_t sr;
	void *buf;
	uint32_t state;
	uint32_t tx;
	uint32_t rx;
	uint32_t rdiff;
} ICollector[MAX_ICOLLECTOR];
static int icollector;
#endif /* USE_ICOLLECTOR */

/*
 * DRV8908 is connected to SPI3 on TXT2
 * The following pins are used:
 * PA15 - SPI3_NSS
 * PE0  - SPI3_SCLK
 * PD10 - SPI3_MISO
 * PD6  - SPI3_MOSI
 * PE14 - SLEEP*
 * PE15 - FAULT*
 */

#define DRVREG(reg) DRV_ ## reg							 /*!< Convert name to enum */
#define RDRV(reg)         ((1<<14) | (DRVREG(reg) << 8)) /*!< Read DRV register */
#define WDRV(reg, value)  ((DRVREG(reg) << 8) | (value)) /*!< Write DRV register */

#if TXT2_BOARD_REVISION == 0
static int DRV8908_old_map = 0;
#endif /* TXT2_BOARD_REVISION == 0 */

#define DRV8908_FAULT_RESET_TIMEOUT 10 /*!< Timeout before reset fault state 10 ms */
static uint32_t DRV8908_ProblemTick = 0;
static int DRV8908_WaitReset = 0;

#if TXT2_BOARD_REVISION == 0
/*
 * Pins on the board:
 * X23 HB2 X24 HB8
 * X25 HB3 X26 HB6
 * X27 HB4 X28 HB7
 * X29 HB5 X30 HB1
 */
static int n_to_hb[] = {
	(2 - 1), (8 - 1),
	(3 - 1), (6 - 1),
	(4 - 1), (7 - 1),
	(5 - 1), (1 - 1)
};
#else
/*
 * Pins on the board:
 * X23 HB2 X24 HB3
 * X25 HB1 X26 HB4
 * X27 HB6 X28 HB8
 * X29 HB5 X30 HB7
 */
static int n_to_hb[] = {
	(2 - 1), (3 - 1),
	(1 - 1), (4 - 1),
	(6 - 1), (8 - 1),
	(5 - 1), (7 - 1)
};
#endif

static uint16_t oldBitsConversion1[256];
static uint16_t oldBitsConversion2[256];

static uint16_t init_sequence[] = { /* Register + value */
	RDRV(IC_STAT),
	RDRV(OCP_STAT_1),
	RDRV(OCP_STAT_2),
	RDRV(OCP_STAT_3),
	RDRV(OLD_STAT_1),
	RDRV(OLD_STAT_2),
	RDRV(OLD_STAT_3),
	RDRV(OP_CTRL_1),
	RDRV(OP_CTRL_2),
	RDRV(OP_CTRL_3),
    /* Temporary disable all outputs */
	WDRV(OP_CTRL_1, 0x00),
	WDRV(OP_CTRL_2, 0x00),
	WDRV(OP_CTRL_3, 0x00),
	/* Disable Open Load Detection */
	//WDRV(OLD_CTRL_1, 0xff),
	//WDRV(OLD_CTRL_2, 0xff),
	WDRV(OLD_CTRL_2, 0xc0), /* No OLD report */
	WDRV(CONFIG_CTRL, 0x01), /* clear faults */
	WDRV(PWM_CTRL_2, 0xf0), /* Disable 4 PWM generators */
	WDRV(FW_CTRL_1, 0x00),
	WDRV(FW_CTRL_2, 0x00),
	WDRV(OCP_OLD_CTRL_1, 0xff), /* Disable OLD to check chip revision */
	RDRV(OCP_OLD_CTRL_1), /* -3 */
	RDRV(IC_STAT),     /* -2 */
	RDRV(CONFIG_CTRL), /* Must be last  -1 */
};

#if TXT2_BOARD_REVISION == 0
/* Initialization sequence for old DRV8908-DRV8912 */
static uint16_t old_init[] = {
	WDRV(OLD_CTRL_2, 0xcf), /* No OLD report */
	WDRV(OLD_CTRL_1, 0xff), /* No OLD report */
	WDRV(CONFIG_CTRL, 0x01), /* clear faults */
    /* All PWM-s already disabled */
	/* Map PWM channels
	 * PWM1 for HB2 HB8
	 * PWM2 for HB3 HB6
	 * PWM3 for HB4 HB7
	 * PWM4 for HB5 HB1
	 */
	/* HB1-HB4 : PWM4 PWM1 PWM2 PWM3 */
	WDRV(PWM_MAP_CTRL_1, (3 << 0) | (0 << 2) | (1 << 4) | (2 << 6)),
	/* HB5-HB8 : PWM4 PWM2 PWM3 PWM1 */
	WDRV(PWM_MAP_CTRL_2, (3 << 0) | (1 << 2) | (2 << 4) | (0 << 6)),
	/* Setup frequency */
	WDRV(PWM_FREQ_CTRL, 0xff), /* setup all PWM frequencies to 2 Khz */
	/* Set 0 to all 8 HB outputs */
	WDRV(PWM_CTRL_1, 0x00), /* Continous mode for all HB */
	WDRV(PWM_CTRL_2, 0x00),/* Enable all PWM generators */
	WDRV(OP_CTRL_1, 0x55), /* enable HBLS */
	WDRV(OP_CTRL_2, 0x55),
	RDRV(IC_STAT)
};
#endif /* TXT2_BOARD_REVISION == 0 */

/* Initialization sequence for new DRV8908 */
static uint16_t new_init[] = {
    /* Disable all PWMs */
	WDRV(PWM_CTRL_2, 0xff),
	WDRV(OP_CTRL_1, 0x00), /* disable all half-bridges */
	WDRV(OP_CTRL_2, 0x00),
	/* Clear all duty values */
	WDRV(PWM_DUTY_CTRL_1, 0x00),
	WDRV(PWM_DUTY_CTRL_2, 0x00),
	WDRV(PWM_DUTY_CTRL_3, 0x00),
	WDRV(PWM_DUTY_CTRL_4, 0x00),
	WDRV(PWM_DUTY_CTRL_5, 0x00),
	WDRV(PWM_DUTY_CTRL_6, 0x00),
	WDRV(PWM_DUTY_CTRL_7, 0x00),
	WDRV(PWM_DUTY_CTRL_8, 0x00),
	/* setup all PWM frequencies to 2 Khz */
	WDRV(PWM_FREQ_CTRL_1, 0xff),
	WDRV(PWM_FREQ_CTRL_2, 0xff),
	/* Map PWMs to outputs */
	/*
	 * Pins on the board:
	 * X23 HB2 X24 HB8
	 * X25 HB3 X26 HB6
	 * X27 HB4 X28 HB7
	 * X29 HB5 X30 HB1
	 */
#if TXT2_BOARD_REVISION == 0
	/* HB1 - O8, HB2 - O1 */
	WDRV(PWM_MAP_CTRL_1, (8 - 1) | ((1 - 1) << 3)),
	/* HB3 - O3, HB4 - O5 */
	WDRV(PWM_MAP_CTRL_2, (3 - 1) | ((5 - 1) << 3)),
	/* HB5 - O7, HB6 - O4 */
	WDRV(PWM_MAP_CTRL_3, (7 - 1) | ((4 - 1) << 3)),
	/* HB7 - O6, HB8 - O2 */
	WDRV(PWM_MAP_CTRL_4, (6 - 1) | ((2 - 1) << 3)),
#else
	/* HB1 - O3, HB2 - O1 */
	WDRV(PWM_MAP_CTRL_1, (3 - 1) | ((1 - 1) << 3)),
	/* HB3 - O2, HB4 - O4 */
	WDRV(PWM_MAP_CTRL_2, (2 - 1) | ((4 - 1) << 3)),
	/* HB5 - O7, HB6 - O5 */
	WDRV(PWM_MAP_CTRL_3, (7 - 1) | ((5 - 1) << 3)),
	/* HB7 - O8, HB8 - O6 */
	WDRV(PWM_MAP_CTRL_4, (8 - 1) | ((6 - 1) << 3)),
#endif
	WDRV(OP_CTRL_1, 0x55), /* enable HBLS */
	WDRV(OP_CTRL_2, 0x55),
	WDRV(PWM_CTRL_2, 0x00), /* enable all PWMs */
	WDRV(SR_CTRL_1, 0xff), /* 2.5 V/mks instead of 0.6 V/mks */
	//WDRV(FW_CTRL_1, 0xff), /* active free-wheeling */
	WDRV(OCP_OLD_CTRL_1, 0x00), /* enable OLD/OCP detection */
	WDRV(OCP_OLD_CTRL_2, 0xc0), /* no nFAULT for OLD */
	WDRV(OCP_OLD_CTRL_3, 0x00),
	WDRV(OCP_OLD_CTRL_4, 0x00), /* disable low current  OLD detection for 1-8 bridges */
	WDRV(OCP_OLD_CTRL_5, 0x00), /* passive OLD operations for 1-8 bridges */
	RDRV(IC_STAT)
};

/* PWM value 0-255 */
static uint16_t rstates[8] = {-1, -1, -1, -1, -1, -1, -1, -1};

#define DRV_89XX_MASK 0x3f /*!< Mask for register number */
static uint16_t drvRegs[DRV_89XX_MASK + 1];

extern uint32_t drvAlertPin; /*!< DRV8908 alert pin value */

/**
 *  DRV8908 state for a next DRV8908 cycle update
 */
struct DRV8908_Context {
	uint8_t pwms[8];	/*!< pwm values for all 8 channels */
	uint8_t hsen[2];	/*!< hsen registers values, 2 bits per channel */
	uint8_t pwm_mode;	/*!< pwm mode for all 8 channels */
};

/* Buffer states:
 * free - free for use
 * ready - ready for output
 * output - output is started
 * busy - new commands can be added
 * free -> busy -> ready -> output -> free
 */
enum{SPIBufferIsFree, SPIBufferBusy, SPIBufferReady, SPIBufferOutput};

/**
 * One cycle (1 ms) buffer for SPI commands
 */
static struct SPICommandBuffer {
	int state; /*!< Buffer state */
	int ipos;  /*!< New position for commands and length */
	int opos;  /*!< New position for output */
	uint16_t commands[sizeof(struct DRV8908_Context) + 10]; /*!< Commands */
} SPICommandBuffer[2];

static struct SPICommandBuffer *pSpiBuf;
static struct SPICommandBuffer *pSpiOBuf;
static int spiIBuf = 0;
static int spiOBuf = 0;

static struct DRV8908_Context prevContext;
static struct DRV8908_Context nextContext = { {0}, {0x55, 0x55}, 0};

#define SPI_RING_SIZE 32 /*!< ring size to compare output-input values */
static uint16_t txring[SPI_RING_SIZE];  /* TX headers    */
static int txpos = 0; /* Ring positions */

static int rxpos = 0;
static uint16_t rxring[SPI_RING_SIZE]; /* Return values */
/*
#define SPI_RX_TEST 64
static uint16_t rxtring[SPI_RX_TEST];
static int rxtpos = 0;
*/
static int rtxflag = 0;

#define NELEMENTS(arr) (sizeof(arr)/sizeof(arr[0]))	/*!< Number of elements in an array */
static uint16_t r[
		NELEMENTS(init_sequence) > NELEMENTS(new_init) ?
				NELEMENTS(init_sequence) : NELEMENTS(new_init)];

/**
 * @brief Initialization GPIO procedure for DRV8908 and MP28167GQ +9V25
 */
static void DRV8908_GPIO_init()
{
	GPIO_InitTypeDef  GPIO_InitStruct;

	for (int i = 0; i < NELEMENTS(drvRegs); i++) {
		drvRegs[i] = 256;
	}

	/* Configure GPIO pins */
	GPIO_InitStruct.Pin = MOTOR_NSLEEP_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_HIGH;
	HAL_GPIO_Init(MOTOR_NSLEEP_GPIO_Port, &GPIO_InitStruct);
	HAL_GPIO_WritePin(MOTOR_NSLEEP_GPIO_Port, MOTOR_NSLEEP_Pin, GPIO_PIN_RESET);

	GPIO_InitStruct.Pin = PWR_BUCK_EN_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
	HAL_GPIO_Init(PWR_BUCK_EN_GPIO_Port, &GPIO_InitStruct);
	/* Switch off external power */
	HAL_GPIO_WritePin(PWR_BUCK_EN_GPIO_Port, PWR_BUCK_EN_Pin, GPIO_PIN_RESET);

    GPIO_InitStruct.Pin = MOTOR_NFAULT_Pin;
    GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
    GPIO_InitStruct.Pull = GPIO_PULLUP;
    GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
    HAL_GPIO_Init(MOTOR_NFAULT_GPIO_Port, &GPIO_InitStruct);

    GPIO_InitStruct.Pin = PWR_BUCK_ALERT_Pin;
    GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
    GPIO_InitStruct.Pull = GPIO_PULLUP;
    GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
    HAL_GPIO_Init(PWR_BUCK_ALERT_GPIO_Port, &GPIO_InitStruct);
}

/* Exported functions */

/**
 * @brief Power down procedure for DRV8908 and MP28167GQ +9V25, switch off 9V25
 */
void DRV8908_Stop()
{
	/* Sleep DRV8908 */
	HAL_GPIO_WritePin(MOTOR_NSLEEP_GPIO_Port, MOTOR_NSLEEP_Pin, GPIO_PIN_RESET);

	/* Switch off external voltage */
	HAL_GPIO_WritePin(PWR_BUCK_EN_GPIO_Port, PWR_BUCK_EN_Pin, GPIO_PIN_RESET);
}

/**
 * @brief Send initialization sequence of DRV8908 frames
 */
static void DRV8908_spi_command(uint16_t *tx, uint16_t *rx, int size)
{
	/* Copy without EOT */
	MODIFY_REG(SPI3->CR2, SPI_CR2_TSIZE, 0);
	SET_BIT(SPI3->CR1, SPI_CR1_SPE);
	SET_BIT(SPI3->CR1, SPI_CR1_CSTART);

	int rs = size, ts = size;
	while (rs || ts) {
		uint32_t sr = SPI3->SR;
		if (sr & SPI_SR_SUSP) {
			/* Clear suspension bit */
			SET_BIT(SPI3->IFCR , SPI_IFCR_SUSPC);
			sr = SPI3->SR;
		}
		if ((rs > 0) && (sr & 0xe001)) {
			if (sr & SPI_SR_RXP) {
				for (int i = 0; rs > 0 && i < PACKET_LENGTH; i++) {
					*rx++ = *((__IO uint16_t *) &SPI3->RXDR);
					rs--;
				}
			} else {
				*rx++ = *((__IO uint16_t *) &SPI3->RXDR);
				rs--;
			}

		}
		if ((ts > 0) && (sr & 2)) {
			for (int i = 0; ts > 0 && i < PACKET_LENGTH; i++) {
				*((__IO uint16_t *)&SPI3->TXDR) = *tx++;
				ts--;
			}
		}
	}
}

/**
 * @brief Read and process one DRV8908 frame
 */
static void DRV8908_Process_Word()
{
	uint32_t sr = SPI3->SR;
	uint16_t word = *((__IO uint16_t *) &SPI3->RXDR);

    //rxtring[rxtpos++] = word;
    //rxtpos &= SPI_RX_TEST - 1;

	if (!word) {
		rtxflag++;
		log_err("zero 0x%lx\n", sr);
	} else if (rxpos == txpos) {
		log_err("rxpos %d txpos %d\n", rxpos, txpos);
		/* TODO */
	} else {
		drvRegs[(txring[rxpos] >> 8) & DRV_89XX_MASK] = word & 0xff;
		drvRegs[DRV_IC_STAT] = (drvRegs[DRV_IC_STAT] & DRV_STAT_SAVE) | ((word >> 8) & DRV_STAT_MASK);
		rxring[rxpos++] = word;
		rxpos &= SPI_RING_SIZE - 1;
		rdiff--;
	}
}

#define DRV8908_send_word(value) pSpiBuf->commands[pSpiBuf->ipos++] = (value) /*!< put word to buffer */

/**
 * @brief Convert all context changes to DRV89xx frames, and send them
 */
static void DRV8908_update_context()
{
	/* Select buffer */
	__disable_irq();
	pSpiBuf = &SPICommandBuffer[spiIBuf];
	if (pSpiBuf->state == SPIBufferOutput) {
		pSpiBuf = &SPICommandBuffer[1 - spiIBuf];
	} else {
		spiIBuf = 1 - spiIBuf; /* For next session */
	}
	pSpiBuf->state = SPIBufferBusy;
	pSpiBuf->ipos = 0;
	__enable_irq();

    /* Fix pwm values */
#if TXT2_BOARD_REVISION == 0
	if (DRV8908_old_map) {
		for (int i = 0; i < 4; i++) {
			/* Old name for DRV_PWM_FREQ_CTRL_1 was DRV_PWM_DUTY_CTRL_1 */
			DRV8908_send_word(((DRV_PWM_FREQ_CTRL_1 + i) << 8) | nextContext.pwms[i]);
			prevContext.pwms[i] = nextContext.pwms[i];
		}
	} else {
#else
	{
#endif
		for (int i = 0; i < 8; i++) {
			DRV8908_send_word(((DRV_PWM_DUTY_CTRL_1 + i) << 8) | nextContext.pwms[i]);
			prevContext.pwms[i] = nextContext.pwms[i];
		}
	}

	/* Fix pwm_mode value */
	DRV8908_send_word((DRV_PWM_CTRL_1 << 8) | nextContext.pwm_mode); /* set PWM mode */
	prevContext.pwm_mode = nextContext.pwm_mode;

	/* Fix hsen values */
	for (int i = 0; i < 2; i++) {
			DRV8908_send_word(((DRV_OP_CTRL_1 + i) << 8) | nextContext.hsen[ i ]);
			prevContext.hsen[i] = nextContext.hsen[i];
	}

	uint16_t stat = drvRegs[DRV_IC_STAT];
	if (stat & DRV_STAT_EVT) {

		if (stat & DRV_STAT_OLD) {
			DRV8908_send_word(RDRV(OLD_STAT_1));
			DRV8908_send_word(RDRV(OLD_STAT_2));
		}

		if (stat & DRV_STAT_OCP) {
			DRV8908_send_word(RDRV(OCP_STAT_1));
			DRV8908_send_word(RDRV(OCP_STAT_2));
		}

		if (DRV8908_WaitReset) {
			uint32_t diff = HAL_GetTick() - DRV8908_ProblemTick;
			if (diff >= DRV8908_FAULT_RESET_TIMEOUT) {
				DRV8908_send_word(WDRV(CONFIG_CTRL, 1)); /* Reset event */
				DRV8908_ProblemTick = HAL_GetTick();
			}
		} else {
			DRV8908_WaitReset = 1;
			DRV8908_ProblemTick = HAL_GetTick();
		}

		if (drvAlertPin) { /* read to check new state and clear it */
			DRV8908_send_word(RDRV(IC_STAT));
		}
	} else {
		DRV8908_WaitReset = 0;
	}

	pSpiBuf->state = SPIBufferReady;
	/* Check/enable interrupts */
	SPI3->IER = SPI_IT_RXP | SPI_IT_TXP | SPI_IT_UDR | SPI_IT_OVR | SPI_IT_MODF | SPI_IT_EOT;

}

/**
 * @brief Copy alert pins and DRV89xx registers to shared memory structure
 * @param f buffer for faults info
 */
void DRV8908_Check_Faults(struct ft_faults *f)
{
	uint16_t stat = drvRegs[DRV_IC_STAT];
	uint8_t flags = stat;

	/* Convert values as required */
	if (stat & DRV_STAT_OLD) {
		f->old = oldBitsConversion1[drvRegs[DRV_OLD_STAT_1] & 0xff]
				| oldBitsConversion2[drvRegs[DRV_OLD_STAT_2] & 0xff];
	} else {
		f->old = 0;
	}
	if (stat & DRV_STAT_OCP) {
		f->ocp = oldBitsConversion1[drvRegs[DRV_OCP_STAT_1] & 0xff]
				| oldBitsConversion2[drvRegs[DRV_OCP_STAT_2] & 0xff];
	} else {
		f->ocp = 0;
	}

	f->flags = flags;
}

#if TXT2_BOARD_REVISION == 0
/**
 * @brief Initalization procedure for old DRV8908 or DRV8912
 */
static void DRV8908_old_init()
{
	DRV8908_old_map = 1;
	DRV8908_spi_command(old_init, r, NELEMENTS(old_init));
}
#endif /* TXT2_BOARD_REVISION == 0 */

/**
 * @brief Initalization procedure for unknown DRV89xx
 */
static void DRV8908_unknown_init()
{
#if TXT2_BOARD_REVISION == 0
	int ninit = NELEMENTS(init_sequence);
	int test = r[ninit - 3] & 0xff; /* OLD_CTRL_1 */

	if (test == 0) { /* Old chip regier map is identified */
		DRV8908_old_init();
		return;
	}
#endif /* TXT2_BOARD_REVISION == 0 */

	/* The new register map is identified */
	DRV8908_spi_command(new_init, r, NELEMENTS(new_init));
}

/**
 * @brief Main initalization procedure for any DRV89xx
 */
void DRV8908_init()
{
	static int wait = 0;
	while(wait);

	uint32_t freq = HAL_RCCEx_GetPeriphCLKFreq(RCC_PERIPHCLK_SPI23);

	log_info("spi3 %lu\n", freq);

	DRV8908_GPIO_init();

	/* Wakeup DRV8908 */
	HAL_GPIO_WritePin(MOTOR_NSLEEP_GPIO_Port, MOTOR_NSLEEP_Pin, GPIO_PIN_SET);

	/* Switch on external voltage */
	HAL_GPIO_WritePin(PWR_BUCK_EN_GPIO_Port, PWR_BUCK_EN_Pin, GPIO_PIN_SET);

	/* Prepare tables */
	uint8_t pins[8];
	int i;
	for (i = 0; i < 8; i++) {
		pins[n_to_hb[i]] = i * 2;
	}

	/* Conversion tables for faults HB1-HB8 -> O1-O8 */
	for (i = 0; i < 256; i++) {
		uint16_t v1 = 0;
		uint16_t v2 = 0;
		int j;
		for (j = 0; j < 4; j++) {
			int s = (i >> (j * 2)) & 3;
			v1 |= s << pins[j + 0];
			v2 |= s << pins[j + 4];
		}
		oldBitsConversion1[i] = v1;
		oldBitsConversion2[i] = v2;
	}

	HAL_Delay(2); /* wakeup time */

	/* Try to reset error if exists */
	HAL_SPI_TransmitReceive(&hspi3, (void*) init_sequence, (void*) r, 1, 1);

	/* Common sequence for old and new DRV89XX chip */
	DRV8908_spi_command(init_sequence, r, NELEMENTS(init_sequence));

	/* Try to select correct map */
	int ninit = NELEMENTS(init_sequence);

	/* Check DRV8908 id */
    int ic_id = (r[ninit - 1] >> 4) & 0x7; /* CONFIG_CTRL */

    switch(ic_id) {
#if TXT2_BOARD_REVISION == 0
	case 0: /* DRV8912 */
		log_info("DRV8912 is detected\n");
		DRV8908_old_init();
		break;
    case 1: /* DRV8910 */
    	log_info("DRV8910 is detected\n");
    	DRV8908_old_init();
    	break;
#endif /* TXT2_BOARD_REVISION == 0 */
    case 2:
    	log_info("DRV8908 is detected\n");
    	DRV8908_unknown_init();
    	break;
    default:
    	log_info("Unknown DRV89xx is detected\n");
    	DRV8908_unknown_init();
    	break;
    }

    /* Go to non EOT mode as recommended in errata */
    SPI3->CR2 = 0;
    /* Master transfer start */
    SET_BIT(SPI3->CR1, SPI_CR1_SPE);
    SET_BIT(SPI3->CR1, SPI_CR1_CSTART);

    initok = 1;
}

#if TXT2_BOARD_REVISION == 0
/**
 * @brief Set PWM procedure for old DRV89xx
 * @param n 0-7 for O1-O8
 * @param pwm new PWM value 0-255
 */
static void DRV8908_set_old_pwm(unsigned int n, uint8_t pwm)
{
    int n1 = n / 2; /* One PWM for two channels */
    int hb = n_to_hb[n];
    int hb4 = hb / 4;

    int shift = (hb & 3) * 2;
    nextContext.hsen[ hb4 ] &= ~(3 << shift);
    nextContext.pwms[n1] = pwm;      /* set PWM value */
    nextContext.pwm_mode |= 1 << hb; /* set PWM mode */
    /* Enable  half bridge */
    nextContext.hsen[ hb4 ] |= 2 << shift;
    rstates[n] = pwm;
}
#endif /* TXT2_BOARD_REVISION == 0 */

/**
 * @brief Set PWM procedure for new DRV89xx
 * @param n 0-7 for O1-O8
 * @param pwm new PWM value 0-255
 */
static void DRV8908_set_new_pwm(unsigned int n, uint8_t pwm)
{
	int hb = n_to_hb[n];
	int shift = (hb & 3) * 2;

	nextContext.pwms[n] = pwm;
	uint8_t pbit = 1 << hb;
	hb /= 4;
	uint8_t hsen = nextContext.hsen[hb] & ~(3 << shift);
	if (pwm) {
		nextContext.pwm_mode |= pbit;
		hsen |= 2 << shift;
	} else {
		nextContext.pwm_mode &= ~pbit;
		hsen |= 1 << shift;
	}
	nextContext.hsen[hb] = hsen;
    rstates[n] = pwm;
}

/**
 * @brief Set PWM procedure for any DRV89xx
 * @param n 0-7 for O1-O8
 * @param pwm new PWM value 0-255
 */
void DRV8908_set_pwm(unsigned int n, uint8_t pwm)
{
    if (n >= 8)
    	return;

    if (rstates[n] == pwm)
    	return;

#if TXT2_BOARD_REVISION == 1
    DRV8908_set_new_pwm(n, pwm);
#else
    if (DRV8908_old_map) {
    	if (n & 1) /* Can't do anything for old chip */
    		return;
    	DRV8908_set_old_pwm(n, pwm);
    } else {
    	DRV8908_set_new_pwm(n, pwm);
    }
#endif
	DRV8908_update_context();
}

#if TXT2_BOARD_REVISION != 1
/**
 * @brief The procedure to set all PWMs for old DRV89xx
 * @param pwms new 8 PWM values 0-512
 */
static void DRV8908_set_old_pwms(uint16_t *pwms, uint8_t motormode)
{
    for (int i = 0; i < FT_MAX_MOTORS; pwms += 2, i++) {
		uint16_t pwm1 = pwms[0];
		uint16_t pwm2 = pwms[1];
		int coast = (motormode & (1 << i));

		/* Convert values to hardware 0-255 form */
		if (pwm1 >= 512)
			pwm1 = 511;
		pwm1 /= 2;

		if (pwm2 >= 512)
			pwm2 = 511;
		pwm2 /= 2;

		/* pwm1 has priority */
		if (pwm1)
			pwm2 = 0;

		/* Simple optimization */
		if (rstates[2 * i + 0] == pwm1 && rstates[2 * i + 1] == pwm2)
			continue;

		int hb1 = n_to_hb[2 * i];
		int hb2 = n_to_hb[2 * i + 1];
		int shift1 = (hb1 & 3) * 2;
		int shift2 = (hb2 & 3) * 2;
		int hb14 = hb1 / 4;
		int hb24 = hb2 / 4;

		nextContext.hsen[hb14] &= ~(3 << shift1);
		nextContext.hsen[hb24] &= ~(3 << shift2);

		if (!coast) {
			if (pwm1 == 0) {
				nextContext.pwms[i] = pwm2;
				nextContext.pwm_mode &= ~(1 << hb1);
				nextContext.hsen[hb14] |= 1 << shift1;
			} else {
				nextContext.pwms[i] = pwm1;
				nextContext.pwm_mode |= 1 << hb1;
				nextContext.hsen[hb14] |= 2 << shift1;
			}

			if (pwm2 == 0) {
				nextContext.pwm_mode &= ~(1 << hb2);
				nextContext.hsen[hb24] |= 1 << shift2;
			} else {
				nextContext.pwm_mode |= 1 << hb2;
				nextContext.hsen[hb24] |= 2 << shift2;
			}

			rstates[2 * i + 0] = pwm1;
			rstates[2 * i + 1] = pwm2;
		} else {
			/* coast mode HS_EN = 0 and LS_EN == 0*/
			nextContext.pwm_mode &= ~((1 << hb2) | (1 << hb1));
			nextContext.pwms[i] = 0;
			rstates[2 * i + 0] = -1;
			rstates[2 * i + 1] = -1;
		}
	}
}
#endif /* TXT2_BOARD_REVISION != 1 */

/**
 * @brief The procedure to set all PWMs for new DRV89xx
 * @param pwms new 8 PWM values 0-512
 * @param motormode motor coast mode
 */
static void DRV8908_set_new_pwms(uint16_t *pwms, uint8_t motormode)
{
	if (pwms[0] | pwms[1]) {
		if (!firstMotorCheck) {
			motorCheckTick = HAL_GetTick();
			firstMotorCheck = 1;
		}
	} else {
		firstMotorCheck = 0;
	}

    for (int i = 0; i < FT_MAX_OUTPUTS; i++) {
		uint16_t pwm = *pwms++;

		if (pwm >= 512)
			pwm = 511;

		pwm /= 2;
		if (rstates[i] == pwm)
			continue;

		int coast = motormode & (1 << (i / 2));
		int hb = n_to_hb[i];
		int shift = (hb & 3) * 2;
		rstates[i] = nextContext.pwms[i] = pwm;
		uint8_t pbit = 1 << hb;
		hb /= 4;
		uint8_t hsen = nextContext.hsen[hb] & ~(3 << shift);

		if (!coast) {
			if (pwm) {
				nextContext.pwm_mode |= pbit;
				hsen |= 2 << shift;
			} else {
				nextContext.pwm_mode &= ~pbit;
				hsen |= 1 << shift;
			}
		} else {
			nextContext.pwm_mode &= ~pbit;
			nextContext.pwms[i] = 0;
			rstates[i] = -1;
		}
		nextContext.hsen[hb] = hsen;
	}
}

static int firstint = 0;
/**
 * @brief The procedure to set all PWMs for any DRV89xx
 * @param pwms new 8 PWM values 0-512
 * @param motormode motor coast mode
 */
void DRV8908_set_pwms(uint16_t *pwms, uint8_t motormode)
{
	firstint = 1;
#if TXT2_BOARD_REVISION == 1
	DRV8908_set_new_pwms(pwms, motormode);
#else
	if (DRV8908_old_map) {
		DRV8908_set_old_pwms(pwms, motormode);
	} else {
		DRV8908_set_new_pwms(pwms, motormode);
	}
#endif
	DRV8908_update_context();
}

/**
 * @brief The procedure to stop motor for any DRV89xx
 * @param n motor number 0, 2, 4 or 6
 */
static void DRV8908_motor_stop_old(unsigned int n)
{
    int hb1 = n_to_hb[n];
    int hb2 = n_to_hb[n + 1];
    int shift1 = (hb1 & 3) * 2;
    int shift2 = (hb2 & 3) * 2;
    int hb14 = hb1 / 4;
    int hb24 = hb2 / 4;

    nextContext.hsen[ hb14 ] &= ~(3 << shift1);
    nextContext.hsen[ hb24 ] &= ~(3 << shift2);

    nextContext.pwm_mode &= ~((1 << hb1) | (1 << hb2)); /* continous mode for both */

    nextContext.hsen[ hb14 ] |= 1 << shift1;
    nextContext.hsen[ hb24 ] |= 1 << shift2;

    rstates[n] = 0;
    rstates[n + 1] = 0;
}

#if TXT2_BOARD_REVISION != 1
/**
 * @brief The procedure to setup motor speed for old DRV89xx
 * @param n motor number 0, 2, 4 or 6
 * @param pwm1 hardware PWM value 0-255 for first pin
 * @param pwm1 hardware PWM value 0-255 for seconf pin
 */
static void DRV8908_motor_speed_old(unsigned int n, uint8_t pwm1, uint8_t pwm2)
{
	int n1 = n / 2; /* One PWM for two channels */

    /* pwm1 or pwm2 must be 0 */
    uint8_t npwm = (pwm1 == 0) ? pwm2 : pwm1;

    if (npwm == 0) {
    	/* Stop case */
    	DRV8908_motor_stop_old(n);
    	return;
    }

    int hb1 = n_to_hb[n];
    int hb2 = n_to_hb[n + 1];
    int shift1 = (hb1 & 3) * 2;
    int shift2 = (hb2 & 3) * 2;
    int hb14 = hb1 / 4;
    int hb24 = hb2 / 4;

    nextContext.hsen[ hb14 ] &= ~(3 << shift1);
    nextContext.hsen[ hb24 ] &= ~(3 << shift2);
    nextContext.pwms[n1] = npwm;

	if (pwm1 == 0) {
		nextContext.pwm_mode |= 1 << hb2;
		nextContext.pwm_mode &= ~(1 << hb1);
		nextContext.hsen[hb14] |= 1 << shift1;
		nextContext.hsen[hb24] |= 2 << shift2;
	} else {
		nextContext.pwm_mode |= 1 << hb1;
		nextContext.pwm_mode &= ~(1 << hb2);
		nextContext.hsen[hb14] |= 2 << shift1;
		nextContext.hsen[hb24] |= 1 << shift2;
	}

	rstates[n] = pwm1;
	rstates[n + 1] = pwm2;
}
#endif /* TXT2_BOARD_REVISION != 1 */

/**
 * @brief The procedure to setup motor speed for new DRV89xx
 * @param n motor number 0, 2, 4 or 6
 * @param pwm1 hardware PWM value 0-255 for first pin
 * @param pwm1 hardware PWM value 0-255 for seconf pin
 */
static void DRV8908_motor_speed_new(unsigned int n, uint8_t pwm1, uint8_t pwm2)
{
    /* pwm1 or pwm2 must be 0 */
    uint8_t npwm = (pwm1 == 0) ? pwm2 : pwm1;

	if (npwm == 0) {
		/* Stop case */
		DRV8908_motor_stop_old(n);
		return;
	}

    int hb1 = n_to_hb[n];
    int hb2 = n_to_hb[n + 1];
    int shift1 = (hb1 & 3) * 2;
    int shift2 = (hb2 & 3) * 2;
    int hb14 = hb1 / 4;
    int hb24 = hb2 / 4;

    nextContext.pwms[n] = pwm1;
    nextContext.pwms[n + 1] = pwm2;

	uint8_t hsen = nextContext.hsen[hb14] & ~(3 << shift1);
	if (pwm1 == 0) {
		nextContext.pwm_mode &= ~(1 << hb1);
		hsen |= 1 << shift1;
	} else {
		nextContext.pwm_mode |= 1 << hb1;
		hsen |= 2 << shift1;
	}
	nextContext.hsen[hb14] = hsen;

	hsen = nextContext.hsen[hb24] & ~(3 << shift2);
	if (pwm2 == 0) {
		nextContext.pwm_mode &= ~(1 << hb2);
		hsen |= 1 << shift2;
	} else {
		nextContext.pwm_mode |= 1 << hb2;
		hsen |= 2 << shift2;
	}
	nextContext.hsen[hb24] = hsen;
	rstates[n] = pwm1;
	rstates[n + 1] = pwm2;
}

/**
 * @brief The procedure to setup motor speed for any DRV89xx
 * @param n motor number 0, 2, 4 or 6
 * @param npwm two PWM values 0-512 for first and second pins
 */
void DRV8908_motor_speed(unsigned int n, ft_pwm_type *npwm)
{
	if (n >= 8)
		return;

	uint16_t pwm1 = npwm[0];
	uint16_t pwm2 = npwm[1];

	if (pwm1 >= 512)
		pwm1 = 511;

	if (pwm2 >= 512)
		pwm2 = 511;

	pwm1 /= 2;
	pwm2 /= 2;

    /* Some optimization */
	if ((pwm1 == rstates[n]) && (pwm2 == rstates[n + 1]))
		return;

#if TXT2_BOARD_REVISION == 1
	DRV8908_motor_speed_new(n, pwm1, pwm2);
#else
	if (DRV8908_old_map) {
		DRV8908_motor_speed_old(n, pwm1, pwm2);
	} else {
		DRV8908_motor_speed_new(n, pwm1, pwm2);
	}
#endif
    DRV8908_update_context();
}

/**
 * @brief The procedure to select next buffer for output in interrupt procedure
 */
static void DRV8908_Select_Buffer()
{
	if (pSpiOBuf == NULL || pSpiOBuf->state != SPIBufferOutput) {
		/* Select new buffer */
		pSpiOBuf = &SPICommandBuffer[spiOBuf];
		if (pSpiOBuf->state != SPIBufferReady) {
			spiOBuf = 1 - spiOBuf;
			pSpiOBuf = &SPICommandBuffer[spiOBuf];
			if (pSpiOBuf->state != SPIBufferReady) {
				/* Only RXP and errors */
				SPI3->IER = SPI_IT_RXP | SPI_IT_UDR | SPI_IT_OVR | SPI_IT_MODF;
				return;
			}
		}
		/* New buffer is selected */
		pSpiOBuf->state = SPIBufferOutput;
		pSpiOBuf->opos = 0;
		tstart = HAL_GetTick();
	}
}

static uint32_t spiint;

/**
 * @brief The SPI interrupt procedure for DRV8908 input/output
 */
void SPI3_IRQHandler(void)
{

	ftstat->spi3counter++;

	if (!initok || !(SPI3->CR1 & SPI_CR1_SPE)) {
		SPI3->IER = 0;
		return; /* errata */
	}

    spiint = HAL_GetTick();

#if USE_ICOLLECTOR
	icollector++;
	icollector &= MAX_ICOLLECTOR - 1;
    ICollector[icollector].itime = HAL_GetTick();
    ICollector[icollector].sr = SPI3->SR;
    ICollector[icollector].buf = pSpiOBuf;
    ICollector[icollector].state = (pSpiOBuf != NULL) ? pSpiOBuf->state : 0x1234;
    ICollector[icollector].tx = 0;
    ICollector[icollector].rx = 0;
    ICollector[icollector].rdiff = rdiff;
#endif /* USE_ICOLLECTOR */

    /* Process SUSP */
	if (SPI3->SR & SPI_SR_SUSP) {
		/* Clear suspension bit */
		log_err("susp 0x%lx\n", SPI3->SR);
		SET_BIT(SPI3->IFCR, SPI_IFCR_SUSPC);
	}
	if (SPI3->SR & (SPI_SR_MODF | SPI_SR_OVR | SPI_SR_UDR)) {
		log_err("sr 0x%lx\n", SPI3->SR);
		if (SPI3->SR & SPI_SR_OVR) {
			SET_BIT(SPI3->IFCR, SPI_IFCR_OVRC); /* TODO: ? */
		}
	}

	/* Process inputs */
	while (SPI3->SR & 0xe001) {
		if (initok) {
	        if (SPI3->SR & SPI_SR_RXP) {
			    for (int i = 0; i < PACKET_LENGTH; i++) {
                    DRV8908_Process_Word();
#if USE_ICOLLECTOR
                    ICollector[icollector].rx++;
#endif /* USE_ICOLLECTOR */
			    }
	        } else {
                DRV8908_Process_Word();
#if USE_ICOLLECTOR
			    ICollector[icollector].rx++;
#endif /* USE_ICOLLECTOR */
	        }
		} else {
			uint16_t word = *((__IO uint16_t *) &SPI3->RXDR);
			(void)word;
		}
	}

	if (!initok) {
		SPI3->IER = 0;
		return; /* errata */
	}

	DRV8908_Select_Buffer();
#if USE_ICOLLECTOR
	ICollector[icollector].buf = pSpiOBuf;
	ICollector[icollector].state = pSpiOBuf->state;
	ICollector[icollector].rdiff = rdiff;
#endif /* USE_ICOLLECTOR */
	if (pSpiOBuf->state != SPIBufferOutput)
		return;

	/* Output state */
	while (SPI3->SR & SPI_SR_TXP) {
		if (pSpiOBuf->opos < pSpiOBuf->ipos) {
			int n = (SPI3->SR & SPI_SR_TXC) ? 8 : PACKET_LENGTH;
			for (int i = 0; i < n; i++) {
				uint16_t word = pSpiOBuf->commands[pSpiOBuf->opos++];
				*((__IO uint16_t *) &SPI3->TXDR) = word;
				txring[txpos++] = word;
				txpos &= SPI_RING_SIZE - 1;
#if USE_ICOLLECTOR
				ICollector[icollector].tx++;
#endif /* USE_ICOLLECTOR */
				rdiff++;


				/* input ? */
				while (SPI3->SR & 0xe001) {
			        if (SPI3->SR & SPI_SR_RXP) {
					    for (int i = 0; i < PACKET_LENGTH; i++) {
                            DRV8908_Process_Word();
#if USE_ICOLLECTOR
                            ICollector[icollector].rx++;
#endif /* USE_ICOLLECTOR */
					    }
			        } else {
                        DRV8908_Process_Word();
#if USE_ICOLLECTOR
					    ICollector[icollector].rx++;
#endif /* USE_ICOLLECTOR */
			        }
				}

				if (pSpiOBuf->opos >= pSpiOBuf->ipos)
					break;
			}
#if USE_ICOLLECTOR
			ICollector[icollector].rdiff = rdiff;
#endif /* USE_ICOLLECTOR */
		} else { /* End of buffer */
			tend = HAL_GetTick();
			if (tend - tstart > tmax) {
				tmax = tend - tstart;
				log_info("tmax %lu\n", tmax);
			}
			pSpiOBuf->state = SPIBufferIsFree;
			DRV8908_Select_Buffer();
#if USE_ICOLLECTOR
			ICollector[icollector].buf = pSpiOBuf;
			ICollector[icollector].state = pSpiOBuf->state;
			ICollector[icollector].rdiff = rdiff;
#endif /* USE_ICOLLECTOR */
			if (pSpiOBuf->state != SPIBufferOutput)
				return;
		}
	}

	if (pSpiOBuf->opos < pSpiOBuf->ipos)
		return;

	pSpiOBuf->state = SPIBufferIsFree;
	DRV8908_Select_Buffer();
#if USE_ICOLLECTOR
	ICollector[icollector].buf = pSpiOBuf;
	ICollector[icollector].state = pSpiOBuf->state;
	ICollector[icollector].rdiff = rdiff;
#endif /* USE_ICOLLECTOR */
}

#if 0
/*
 * Unused
 */
void DRV8908Check()
{
	if (!firstint)
		return;

	uint32_t ctick = HAL_GetTick();

	if (ctick - spiint < 1) {
		return;
	}
	/* something wrong */
	__disable_irq();
	if (SPI3->SR & SPI_SR_SUSP) {
		SET_BIT(SPI3->IFCR , SPI_IFCR_SUSPC); /* Where is interrupt ? */
	} else if (SPI3->SR){
	    DRV8908_Interrupt();
	} else {
		uint32_t sr1 = SPI3->SR;
		if (!sr1)
			sr1 = SPI3->SR;
		if (sr1) {
			__enable_irq();
			return;
		}

	    /* Stop SPI */

		SPI3->CR1 &= ~SPI_CR1_SPE;
		/* Fix rings */
		txpos = 0;
		rxpos = 0;
		rdiff = 0;
		timeouts++;
		/* Restart SPI */
		SPI3->CR1 |= SPI_CR1_SPE;
		SPI3->CR1 |= SPI_CR1_CSTART;
		sr1 = SPI3->SR;
		if (sr1) {
			DRV8908_Interrupt();
		}
	}
	__enable_irq();
}
#endif

