/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file HWFDCAN.c
 * @brief FDCAN procedures, initialization and communication
 */

#include "HWFDCAN.h"
#include "main.h"
#include "ExtEngine.h"
#include "ftcalib.h"

#include "cmsis_os.h"
#include <semphr.h>
#include <string.h>
#include <task.h>

extern struct ftcalib ftcalib;	/*!< Common calibration values */

uint32_t slavePacketTime[FT_MAX_TXT * 2 + 4]; /*!< Timeout for errata */

/*
 * CAN matrix 4 slots for each basic cycle
 *
 */
#define TM0   ftcalib.tm[0] /*!< Time marker 0 */
#define TM1   ftcalib.tm[1] /*!< Time marker 1 */
#define TM2   ftcalib.tm[2] /*!< Time marker 2, unused */
#define TM3   ftcalib.tm[3] /*!< Time marker 3 */
#define TM4   ftcalib.tm[4] /*!< Time marker 4 */

extern FDCAN_HandleTypeDef hfdcan1;			/*!< FDCAN1 HAL handle */
static FDCAN_FilterTypeDef sFilterConfig;
static FDCAN_TxHeaderTypeDef TxHeader;
static FDCAN_TxEventFifoTypeDef txevent;


static int fdcanMode = 0;
static int fdcanBitrate;

/* FDCAN_DLC_BYTES_0, FDCAN_DLC_BYTES_1,
   FDCAN_DLC_BYTES_2, FDCAN_DLC_BYTES_3, FDCAN_DLC_BYTES_4,
   FDCAN_DLC_BYTES_5, FDCAN_DLC_BYTES_6, FDCAN_DLC_BYTES_7,
   FDCAN_DLC_BYTES_8

   FDCAN_DLC_BYTES_12, FDCAN_DLC_BYTES_16, FDCAN_DLC_BYTES_20, FDCAN_DLC_BYTES_24

   FDCAN_DLC_BYTES_32, FDCAN_DLC_BYTES_48, FDCAN_DLC_BYTES_64
   */
#define DLC_length(n) ((((n) <= 8) ? (n) : ((n) <= 24 ? ((n) + 3) / 4 + 6 \
													  : ((n) + 15) / 16 + 11)) << 16) \
														/*!< Macro to convert length -> DLC */

static uint8_t TxDataCalib[] = {0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55};
static uint8_t TxDataMarker[8] = {0, 0, 0, 0, 0x12, 0x34, 0x56, 0x78};

static union {
	uint8_t TxData[64];
	TXT2_ext_command command;
} TxCommand;

#define IE_ENABLE_FLAGS (FDCAN_IE_BOE | FDCAN_IE_DRXE | FDCAN_IE_RF0WE | FDCAN_IE_RF0NE \
		| FDCAN_IE_RF1WE | FDCAN_IE_RF1NE | FDCAN_IE_TCE | FDCAN_IE_TEFWE | FDCAN_IE_TEFNE \
		| FDCAN_IE_TOOE) /*!< flags for FDCAN IE Register */
#define TTIE_ENABLE_FLAGS (FDCAN_TTIE_AWE | FDCAN_TTIE_CERE | FDCAN_TTIE_IWTE | FDCAN_TTIE_SE2E \
		| FDCAN_TTIE_TXOE | FDCAN_TTIE_WTE) /*!< flags for FDCAN TTIE Register */

static FDCAN_RxHeaderTypeDef CanRxHeader;
static uint8_t CanRxData[64];

/** Structure to define parameters for different bitrates */
static const struct FDCANParameters {
	int bitrate;
	uint16_t NominalTimeSeg1;
	uint16_t NominalTimeSeg2;
	uint16_t DataPrescaler;
	uint8_t  DataTimeSeg1;
	uint8_t  DataTimeSeg2;
} fdcanParameters[] = {
	{ 125000, 0xa7,0x18, 0x6, 0x1b,0x4},
	{ 250000, 0x53,0x0b, 0x3, 0x1b,0x4},
	{ 500000, 0x29,0x07, 0x2, 0x14,0x3},
	{1000000, 0x11,0x06, 0x1, 0x11,0x6},
	{2000000, 0x11,0x06, 0x1, 0x08,0x3},
	{4000000, 0x11,0x06, 0x2, 0x01,0x1},
	{5000000, 0x11,0x06, 0x1, 0x02,0x1}
};

#define USE_TTCAN_COLLECTOR 0 /*!< Switch on TTCAN debug */
#if USE_TTCAN_COLLECTOR
#define TTCAN_COLLECTOR_MAX 256
static struct {
	uint32_t tick;
	uint32_t ttir;
	uint32_t cccr;
	uint32_t fifo;
	uint32_t masterSilentMode;
	uint32_t masterSilentTick;
} ttcollector[TTCAN_COLLECTOR_MAX];
static int icollector;
#endif

static StaticSemaphore_t hwFDSemaphore;
static SemaphoreHandle_t hwFDEventHandle;

/**
 * @brief Init FDCAN GPIO pins: standby and sleep
 */
void HWFDCAN_init()
{
	GPIO_InitTypeDef  GPIO_InitStruct;

	GPIO_InitStruct.Pin = CAN_STANDBY_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
	GPIO_InitStruct.Pull = GPIO_PULLDOWN;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
	HAL_GPIO_Init(CAN_STANDBY_GPIO_Port, &GPIO_InitStruct);
	HAL_GPIO_WritePin(CAN_STANDBY_GPIO_Port, CAN_STANDBY_Pin, GPIO_PIN_RESET);


	GPIO_InitStruct.Pin = CAN_TERM_Pin;
	GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
	HAL_GPIO_Init(CAN_TERM_GPIO_Port, &GPIO_InitStruct);

	hwFDEventHandle = xSemaphoreCreateBinaryStatic(&hwFDSemaphore);
}

/**
 * @brief Enable FDCAN interrupts
 */
static void HWFDCAN_Enable_Interrupts()
{
	FDCAN1->IE = IE_ENABLE_FLAGS;
	TTFDCAN1->TTIE = TTIE_ENABLE_FLAGS;
}

/**
 * @brief Select FDCAN terminator mode
 * @param mode new terminator mode: 0 - no, 1 - use terminator
 */
void HWFDCAN_set_terminator(int mode)
{
	HAL_GPIO_WritePin(CAN_TERM_GPIO_Port, CAN_TERM_Pin, mode);
	ftstat->fdterm = mode;
}

/**
 * @brief Init master TTCAN mode
 */

static void HWFDCAN_TTCAN_master()
{
	FDCAN_TT_ConfigTypeDef TTParams;
	TTParams.OperationMode = FDCAN_TT_COMMUNICATION_LEVEL2;
	TTParams.GapEnable = 0;
	TTParams.TimeMaster = FDCAN_TT_POTENTIAL_MASTER;
	TTParams.SyncDevLimit = 3;
	TTParams.InitRefTrigOffset = 127;
	TTParams.ExternalClkSync = FDCAN_TT_EXT_CLK_SYNC_DISABLE;
	TTParams.AppWdgLimit = 0;
	TTParams.GlobalTimeFilter = 0;
	TTParams.ClockCalibration = FDCAN_TT_AUTO_CLK_CALIB_DISABLE;
	TTParams.EvtTrigPolarity = 0;
	TTParams.BasicCyclesNbr = 9 + 9; /* 10 TXT2 devices */
	TTParams.CycleStartSync = 0;
	TTParams.TxEnableWindow = 3;
	TTParams.ExpTxTrigNbr = 90; //(1 + 2 * 9) * 2;
	TTParams.TURNumerator = ftcalib.cnr;
	TTParams.TURDenominator = ftcalib.dnr;
	TTParams.TriggerMemoryNbr = 64;
	TTParams.StopWatchTrigSel = 0;
	TTParams.EventTrigSel = 0;

	HAL_FDCAN_TT_ConfigOperation(&hfdcan1, &TTParams);
	HAL_FDCAN_ConfigTimestampCounter(&hfdcan1, 1 - 1);
	HAL_FDCAN_EnableTimestampCounter(&hfdcan1, 1);

	HAL_FDCAN_TT_ConfigReferenceMessage(&hfdcan1, FDCAN_STANDARD_ID,
			FT_CAN_MARKER_ID, 0x80000001);

	FDCAN_TriggerTypeDef TriggerConfig;

	TriggerConfig.TriggerIndex = 0;

	TriggerConfig.TimeMark = TM0;
	TriggerConfig.RepeatFactor = FDCAN_TT_REPEAT_EVERY_32ND_CYCLE;
	TriggerConfig.StartCycle = 0;
	TriggerConfig.TmEventInt = 0;
	TriggerConfig.TmEventExt = 0;
	TriggerConfig.TriggerType = FDCAN_TT_RX_TRIGGER;
	TriggerConfig.FilterType = FDCAN_EXTENDED_ID; /* Nslave */
	TriggerConfig.TxBufferIndex = 0;
	TriggerConfig.FilterIndex = 0;

	HAL_FDCAN_TT_ConfigTrigger(&hfdcan1, &TriggerConfig);

	for (int i = 1; i <= 9 * 2; i++) {
		TriggerConfig.TriggerIndex++;
		TriggerConfig.TimeMark = TM0;
		TriggerConfig.RepeatFactor = FDCAN_TT_REPEAT_EVERY_32ND_CYCLE;
		TriggerConfig.StartCycle++;
		TriggerConfig.TmEventInt = 0;
		TriggerConfig.TmEventExt = 0;
		TriggerConfig.TriggerType = FDCAN_TT_RX_TRIGGER;
		TriggerConfig.FilterType = FDCAN_STANDARD_ID; /* Values */
		TriggerConfig.TxBufferIndex = 0;
		TriggerConfig.FilterIndex = 0;

		HAL_FDCAN_TT_ConfigTrigger(&hfdcan1, &TriggerConfig);
	}

	TriggerConfig.TriggerIndex++;
	TriggerConfig.TimeMark = TM1;
	TriggerConfig.RepeatFactor = FDCAN_TT_REPEAT_EVERY_32ND_CYCLE;
	TriggerConfig.StartCycle = 0;
	TriggerConfig.TriggerType = FDCAN_TT_TX_TRIGGER_CONTINUOUS;
	TriggerConfig.FilterType = FDCAN_EXTENDED_ID;
	TriggerConfig.TxBufferIndex = FT_CAN_TX_CALIBRATION; /* Calibration */
	TriggerConfig.FilterIndex = 0;

	HAL_FDCAN_TT_ConfigTrigger(&hfdcan1, &TriggerConfig);

	for (int i = 1; i <= 9; i++) {
		TriggerConfig.TriggerIndex++;
		TriggerConfig.TimeMark = TM1;
		TriggerConfig.RepeatFactor = FDCAN_TT_REPEAT_EVERY_32ND_CYCLE;
		TriggerConfig.StartCycle++;
		TriggerConfig.TriggerType = FDCAN_TT_TX_TRIGGER_SINGLE;
		TriggerConfig.FilterType = FDCAN_STANDARD_ID;
		TriggerConfig.TxBufferIndex = FT_CAN_TX_SADDR(i); /* saddr or config */
		TriggerConfig.FilterIndex = 0;

		HAL_FDCAN_TT_ConfigTrigger(&hfdcan1, &TriggerConfig);
	}

	for (int i = 1; i <= 9; i++) {
		TriggerConfig.TriggerIndex++;
		TriggerConfig.TimeMark = TM1;
		TriggerConfig.RepeatFactor = FDCAN_TT_REPEAT_EVERY_32ND_CYCLE;
		TriggerConfig.StartCycle++;
		TriggerConfig.TriggerType = FDCAN_TT_TX_TRIGGER_SINGLE;
		TriggerConfig.FilterType = FDCAN_STANDARD_ID;
		TriggerConfig.TxBufferIndex = FT_CAN_TX_OP(i); /* op */
		TriggerConfig.FilterIndex = 0;

		HAL_FDCAN_TT_ConfigTrigger(&hfdcan1, &TriggerConfig);
	}

	TriggerConfig.TriggerIndex++;
	TriggerConfig.TimeMark = TM3;
	TriggerConfig.RepeatFactor = FDCAN_TT_REPEAT_EVERY_CYCLE;
	TriggerConfig.StartCycle = 0;
	TriggerConfig.TmEventInt = 0;
	TriggerConfig.TmEventExt = 0;
	TriggerConfig.TriggerType = FDCAN_TT_TX_REF_TRIGGER;
	TriggerConfig.FilterType = FDCAN_STANDARD_ID;
	TriggerConfig.TxBufferIndex = 0;
	TriggerConfig.FilterIndex = 0;

	HAL_FDCAN_TT_ConfigTrigger(&hfdcan1, &TriggerConfig);

	TriggerConfig.TriggerIndex++;
	TriggerConfig.TimeMark = TM4;
	TriggerConfig.TriggerType = FDCAN_TT_WATCH_TRIGGER;

	HAL_FDCAN_TT_ConfigTrigger(&hfdcan1, &TriggerConfig);

	TriggerConfig.TriggerIndex++;
	TriggerConfig.TriggerType = FDCAN_TT_END_OF_LIST;

	HAL_FDCAN_TT_ConfigTrigger(&hfdcan1, &TriggerConfig);
}

/**
 * @brief Init slave TTCAN mode
 */
static void HWFDCAN_TTCAN_slave()
{
	FDCAN_TT_ConfigTypeDef TTParams;
	TTParams.OperationMode = FDCAN_TT_COMMUNICATION_LEVEL2;
	TTParams.GapEnable = 0;
	TTParams.TimeMaster = FDCAN_TT_SLAVE;
	TTParams.SyncDevLimit = 3;
	TTParams.InitRefTrigOffset = 0;
	TTParams.ExternalClkSync = FDCAN_TT_EXT_CLK_SYNC_ENABLE;
	TTParams.AppWdgLimit = 0;
	TTParams.GlobalTimeFilter = 0;
	TTParams.ClockCalibration = FDCAN_TT_AUTO_CLK_CALIB_ENABLE;
	TTParams.EvtTrigPolarity = 0;
	TTParams.BasicCyclesNbr = 9; /* 10 TXT2 devices, unused */
	TTParams.CycleStartSync = 0;
	TTParams.TxEnableWindow = 3;
	TTParams.ExpTxTrigNbr = 1 + 2 * 9;
	TTParams.TURNumerator = ftcalib.cnr;
	TTParams.TURDenominator = ftcalib.dnr;
	TTParams.TriggerMemoryNbr = 64;
	TTParams.StopWatchTrigSel = 0;
	TTParams.EventTrigSel = 0;

	HAL_FDCAN_TT_ConfigOperation(&hfdcan1, &TTParams);
	HAL_FDCAN_ConfigTimestampCounter(&hfdcan1, 1 - 1);
	HAL_FDCAN_EnableTimestampCounter(&hfdcan1, 1);

	HAL_FDCAN_TT_ConfigReferenceMessage(&hfdcan1, FDCAN_STANDARD_ID, FT_CAN_MARKER_ID, 0);

	FDCAN_TriggerTypeDef TriggerConfig;

	TriggerConfig.TriggerIndex = 0;

	TriggerConfig.TimeMark = TM0;
	TriggerConfig.RepeatFactor = FDCAN_TT_REPEAT_EVERY_32ND_CYCLE;
	TriggerConfig.StartCycle = 0;
	TriggerConfig.TmEventInt = 0;
	TriggerConfig.TmEventExt = 0;
	TriggerConfig.TriggerType = FDCAN_TT_TX_TRIGGER_ARBITRATION;
	TriggerConfig.FilterType = FDCAN_EXTENDED_ID; /* Nslave */
	TriggerConfig.TxBufferIndex = FT_CAN_TX_DISCOVERY;
	TriggerConfig.FilterIndex = 0;

	HAL_FDCAN_TT_ConfigTrigger(&hfdcan1, &TriggerConfig);

	for (int i = 1; i <= 9; i++) {
		TriggerConfig.TriggerIndex++;
		TriggerConfig.TimeMark = TM0;
		TriggerConfig.RepeatFactor = FDCAN_TT_REPEAT_EVERY_32ND_CYCLE;
		TriggerConfig.StartCycle++;
		TriggerConfig.TriggerType = FDCAN_TT_TX_TRIGGER_ARBITRATION; //FDCAN_TT_TX_TRIGGER_SINGLE;
		TriggerConfig.FilterType = FDCAN_STANDARD_ID;
		TriggerConfig.TxBufferIndex = FT_CAN_TX_VALUES(i); /* info, values */
		TriggerConfig.FilterIndex = 0;

		HAL_FDCAN_TT_ConfigTrigger(&hfdcan1, &TriggerConfig);
	}

	for (int i = 1; i <= 9; i++) {
		TriggerConfig.TriggerIndex++;
		TriggerConfig.TimeMark = TM0;
		TriggerConfig.RepeatFactor = FDCAN_TT_REPEAT_EVERY_32ND_CYCLE;
		TriggerConfig.StartCycle++;
		TriggerConfig.TriggerType = FDCAN_TT_TX_TRIGGER_ARBITRATION; //FDCAN_TT_TX_TRIGGER_SINGLE;
		TriggerConfig.FilterType = FDCAN_STANDARD_ID;
		TriggerConfig.TxBufferIndex = FT_CAN_TX_VALUES(i); /* info, values */
		TriggerConfig.FilterIndex = 0;

		HAL_FDCAN_TT_ConfigTrigger(&hfdcan1, &TriggerConfig);
	}

	TriggerConfig.TriggerIndex++;
	TriggerConfig.TimeMark = TM1;
	TriggerConfig.RepeatFactor = FDCAN_TT_REPEAT_EVERY_CYCLE;
	TriggerConfig.StartCycle = 0;
	TriggerConfig.TmEventInt = 0;
	TriggerConfig.TmEventExt = 0;
	TriggerConfig.TriggerType = FDCAN_TT_RX_TRIGGER;
	TriggerConfig.FilterType = FDCAN_STANDARD_ID; /* Config */
	TriggerConfig.TxBufferIndex = 0;
	TriggerConfig.FilterIndex = 0;

	HAL_FDCAN_TT_ConfigTrigger(&hfdcan1, &TriggerConfig);

	TriggerConfig.TriggerIndex++;
	TriggerConfig.TimeMark = TM4;
	TriggerConfig.RepeatFactor = FDCAN_TT_REPEAT_EVERY_CYCLE;
	TriggerConfig.StartCycle = 0;
	TriggerConfig.TmEventInt = 0;
	TriggerConfig.TmEventExt = 0;
	TriggerConfig.TriggerType = FDCAN_TT_WATCH_TRIGGER;
	TriggerConfig.FilterType = FDCAN_STANDARD_ID;
	TriggerConfig.TxBufferIndex = 0;
	TriggerConfig.FilterIndex = 0;

	HAL_FDCAN_TT_ConfigTrigger(&hfdcan1, &TriggerConfig);

	TriggerConfig.TriggerIndex++;
	TriggerConfig.TriggerType = FDCAN_TT_END_OF_LIST;

	HAL_FDCAN_TT_ConfigTrigger(&hfdcan1, &TriggerConfig);
}

/**
 * @brief Fix FDCAN register for bitrate
 * @param p new values for FDCAN registers
 */
static void HWFDCAN_fix_parameters(const struct FDCANParameters *p)
{
	FDCAN_ClkCalUnitTypeDef sCcuConfig = {0};

	HAL_FDCAN_DeInit(&hfdcan1);

	hfdcan1.Init.FrameFormat = FDCAN_FRAME_FD_BRS;
	hfdcan1.Init.NominalPrescaler = 1;
	hfdcan1.Init.NominalSyncJumpWidth = 1;
	hfdcan1.Init.NominalTimeSeg1 = p->NominalTimeSeg1;
	hfdcan1.Init.NominalTimeSeg2 = p->NominalTimeSeg2;
	hfdcan1.Init.DataPrescaler = p->DataPrescaler;
	hfdcan1.Init.DataSyncJumpWidth = 1;
	hfdcan1.Init.DataTimeSeg1 = p->DataTimeSeg1;
	hfdcan1.Init.DataTimeSeg2 = p->DataTimeSeg2;

	HAL_FDCAN_Init(&hfdcan1);

	sCcuConfig.ClockDivider = 0;
	if (1 || fdcanMode == ft_txt_master) {
		sCcuConfig.ClockCalibration = FDCAN_CLOCK_CALIBRATION_DISABLE;
	} else {
		sCcuConfig.ClockCalibration = FDCAN_CLOCK_CALIBRATION_ENABLE;
	}
	sCcuConfig.MinOscClkPeriods = 4;
	sCcuConfig.CalFieldLength = FDCAN_CALIB_FIELD_LENGTH_32;
	sCcuConfig.TimeQuantaPerBitTime = 4;
	HAL_FDCAN_ConfigClockCalibration(&hfdcan1, &sCcuConfig);
}

/**
 * @brief Selected new FDCAN mode
 * @param mode new mode: 0 - single, 1 - master, 2 - slave
 */
void HWFDCAN_set_mode(int mode)
{
	fdcanMode = mode;
}

/**
 * @brief Selected new FDCAN bitrate
 * @param bitrate new bitrate (hz), for example, 1000000
 */
void HWFDCAN_set_bitrate(int bitrate)
{
	if (fdcanMode)
		return;

	for (int i = 0; i < sizeof(fdcanParameters) / sizeof(fdcanParameters[0]); i++) {
		if (fdcanParameters[i].bitrate == bitrate) {
			HWFDCAN_fix_parameters(&fdcanParameters[i]);
			ftstat->bitrate = fdcanBitrate = bitrate;
			return;
		}
	}
}

/**
 * @brief Add new selected address to filter for slave
 * @param canAddr CAN id
 */
void HWFDCAN_AddAddressToFilter(int canAddr)
{
	sFilterConfig.IdType = FDCAN_STANDARD_ID;
	sFilterConfig.FilterIndex = 0;
	sFilterConfig.FilterType = FDCAN_FILTER_DUAL;
	sFilterConfig.FilterConfig = FDCAN_FILTER_TO_RXFIFO0;
	sFilterConfig.FilterID1 = FT_CAN_DISCOVERY_ID;
	sFilterConfig.FilterID2 = FT_CAN_BASE_ID + canAddr * 2;
	sFilterConfig.IsCalibrationMsg = 0;

	HAL_FDCAN_ConfigFilter(&hfdcan1, &sFilterConfig);
}

#if 0 /* unused */
/**
 * @brief Send packet procedure
 * @param packet packet buffer
 * @param size buffer size
 * @param addr CAN id
 */
void HWFDCAN_send_packet(uint8_t *packet, int size, int addr)
{
	if (size > 64) {
		log_info("Illegal CAN size %d\n", size);
		return;
	}
	TxHeader.Identifier = addr;
	TxHeader.IdType = (addr & -(1 << 11)) ? FDCAN_EXTENDED_ID : FDCAN_STANDARD_ID;
    TxHeader.TxFrameType = FDCAN_DATA_FRAME;
    TxHeader.DataLength = DLC_length(size);
    TxHeader.ErrorStateIndicator = FDCAN_ESI_ACTIVE;
    TxHeader.BitRateSwitch = (fdcanBitrate > 1000000) ? FDCAN_BRS_ON : FDCAN_BRS_OFF;
    TxHeader.FDFormat = FDCAN_FD_CAN;
    TxHeader.TxEventFifoControl = FDCAN_STORE_TX_EVENTS;
    TxHeader.MessageMarker = addr & 0xff;
    HAL_FDCAN_AddMessageToTxFifoQ(&hfdcan1, &TxHeader, packet);
    ftstat->cantx++;
}
#endif

/**
 * @brief Send packet procedure
 * @param packet packet buffer
 * @param size buffer size
 * @param addr CAN id
 */
void HWFDCAN_set_buffer(uint8_t *packet, int size, int addr, int nbuffer)
{
	if (size > 64) {
		//log_info("Illegal CAN size %d\n", size);
		return;
	}
	TxHeader.Identifier = addr;
	TxHeader.IdType = (addr & -(1 << 11)) ? FDCAN_EXTENDED_ID : FDCAN_STANDARD_ID;
    TxHeader.TxFrameType = FDCAN_DATA_FRAME;
    TxHeader.DataLength = DLC_length(size);
    TxHeader.ErrorStateIndicator = FDCAN_ESI_ACTIVE;
    TxHeader.BitRateSwitch = (fdcanBitrate > 1000000) ? FDCAN_BRS_ON : FDCAN_BRS_OFF;
    TxHeader.FDFormat = FDCAN_FD_CAN;
    TxHeader.TxEventFifoControl = FDCAN_STORE_TX_EVENTS;
    TxHeader.MessageMarker = nbuffer;
    HAL_FDCAN_AddMessageToTxBuffer(&hfdcan1, &TxHeader, packet, 1 << nbuffer);
    /*HAL_FDCAN_EnableTxBufferRequest(&hfdcan1, 1 << nbuffer);*/
    FDCAN1->TXBAR = 1 << nbuffer;
    slavePacketTime[nbuffer] = HAL_GetTick();
    ftstat->cantx++;
}

/**
 * @brief Main Master filter procedure
 * Filters:
 * master: base + 1 for discovery input/output (FIFO0)
 *         base + odd for input FIFO0
 *         extended for initial discovery from slave
 */
void HWFDCAN_ConfigureMasterFilter()
{
	/* Initial discovery message from slave */
	/* Normal message from slave/master */
	sFilterConfig.IdType = FDCAN_EXTENDED_ID;
	sFilterConfig.FilterIndex = 0;
	sFilterConfig.FilterType = FDCAN_FILTER_MASK;
	sFilterConfig.FilterConfig = FDCAN_FILTER_TO_RXFIFO1;
	sFilterConfig.FilterID1 = 0;
	sFilterConfig.FilterID2 = 0;

	HAL_FDCAN_ConfigFilter(&hfdcan1, &sFilterConfig);

	/* Normal message from slave/master */
	sFilterConfig.IdType = FDCAN_STANDARD_ID;
	sFilterConfig.FilterIndex = 0;
	sFilterConfig.FilterType = FDCAN_FILTER_MASK;
	sFilterConfig.FilterConfig = FDCAN_FILTER_TO_RXFIFO0;
	sFilterConfig.FilterID1 = FT_CAN_DISCOVERY_ID;
	sFilterConfig.FilterID2 = (1 | -FT_CAN_ID_RANGE) & 0x7ff;

	HAL_FDCAN_ConfigFilter(&hfdcan1, &sFilterConfig);

	/* Reject all other packets */
	sFilterConfig.FilterIndex++;
	sFilterConfig.FilterType = FDCAN_FILTER_MASK;
	sFilterConfig.FilterConfig = FDCAN_FILTER_REJECT;
	sFilterConfig.FilterID1 = 0;
	sFilterConfig.FilterID2 = 0;

	HAL_FDCAN_ConfigFilter(&hfdcan1, &sFilterConfig);
}

/**
 * @brief Master restart procedure
 * @param MON value
 */
static void HWFDCAN_Master_Restart(uint32_t mon)
{
	SET_BIT(FDCAN1->CCCR, FDCAN_CCCR_INIT);
	while ((FDCAN1->CCCR & FDCAN_CCCR_INIT) == 0U);
	SET_BIT(FDCAN1->CCCR, FDCAN_CCCR_CCE);
	MODIFY_REG(FDCAN1->CCCR, FDCAN_CCCR_MON | FDCAN_CCCR_ASM, mon);
	CLEAR_BIT(FDCAN1->CCCR, FDCAN_CCCR_CCE);
	CLEAR_BIT(FDCAN1->CCCR, FDCAN_CCCR_INIT);
}

/**
 * @brief Main Master procedure to receive and produce FDCAN packages
 */
static void HWFDCAN_Master()
{
	uint32_t stick = 0;

	SET_BIT(FDCAN1->CCCR, FDCAN_CCCR_DAR); /* TODO: Use as common later */
	HWFDCAN_ConfigureMasterFilter();

	/* Prepare marker package */
	TxHeader.Identifier = FT_CAN_MARKER_ID;
	TxHeader.IdType = FDCAN_STANDARD_ID;
	TxHeader.TxFrameType = FDCAN_DATA_FRAME;
	TxHeader.DataLength = FDCAN_DLC_BYTES_8;
	TxHeader.ErrorStateIndicator = FDCAN_ESI_ACTIVE;
	TxHeader.BitRateSwitch = FDCAN_BRS_OFF;
	TxHeader.FDFormat = FDCAN_CLASSIC_CAN;
	TxHeader.TxEventFifoControl = FDCAN_NO_TX_EVENTS;
	TxHeader.MessageMarker = FT_CAN_TXN_MARKER;

	HAL_FDCAN_AddMessageToTxBuffer(&hfdcan1, &TxHeader, TxDataMarker, FT_CAN_TX_MARKER);

	/* Calibration */
	/* Configure calibration message */
	TxHeader.Identifier = 0x555;
	TxHeader.IdType = FDCAN_STANDARD_ID;
	TxHeader.TxFrameType = FDCAN_DATA_FRAME;
	TxHeader.DataLength = FDCAN_DLC_BYTES_8;
	TxHeader.ErrorStateIndicator = FDCAN_ESI_ACTIVE;
	TxHeader.BitRateSwitch = FDCAN_BRS_OFF;
	TxHeader.FDFormat = FDCAN_CLASSIC_CAN;
	TxHeader.TxEventFifoControl = FDCAN_NO_TX_EVENTS;
	TxHeader.MessageMarker = FT_CAN_TXN_CALIBRATION;

	HAL_FDCAN_AddMessageToTxBuffer(&hfdcan1, &TxHeader, TxDataCalib, FT_CAN_TX_CALIBRATION);

	HWFDCAN_TTCAN_master();
	HAL_FDCAN_TT_SetGlobalTime(&hfdcan1, 0);
	HAL_FDCAN_Start(&hfdcan1);
	/* HAL_FDCAN_EnableTxBufferRequest(&hfdcan1, FT_CAN_TX_CALIBRATION | FT_CAN_TX_DISCOVERY); */
	FDCAN1->TXBAR = FT_CAN_TX_CALIBRATION | FT_CAN_TX_DISCOVERY;
	stick = master_tick = HAL_GetTick();

	uint32_t mstate = 0; /* 0 - mon, 1 - non mon */
	HWFDCAN_Enable_Interrupts();

	while(1) {
		xSemaphoreTake(hwFDEventHandle, 10);
		uint32_t mtick = HAL_GetTick();
		uint32_t tvalue = mtick - master_tick;

		/* Fix 0 packet */
		uint32_t *TxAddress = (uint32_t *) (hfdcan1.msgRam.TxBufferSA
				+ (FT_CAN_TXN_MARKER * hfdcan1.Init.TxElmtSize * 4U));
		/* 2 contol words + 1 for marker packet */
		TxAddress[3] = tvalue;

#if USE_TTCAN_COLLECTOR
		icollector = (icollector + 1) & (TTCAN_COLLECTOR_MAX - 1);
		ttcollector[icollector].tick = HAL_GetTick();
		ttcollector[icollector].ttir = TTFDCAN1->TTIR;
		ttcollector[icollector].cccr = FDCAN1->CCCR;
		ttcollector[icollector].fifo = 2;
		ttcollector[icollector].masterSilentMode = masterSilentMode;
		ttcollector[icollector].masterSilentTick = masterSilentTick;
#endif
		if (masterSilentMode && (HAL_GetTick() - masterSilentTick) > 1000) { /* 1 sec */
			masterSilentMode = 0;
		}

		uint32_t ok = 1;

		/* Check state, restart if required */
		if (TTFDCAN1->TTIR & FDCAN_TTIR_IWT) { /* No responce */
			/* Don't do it very quikly */
			TTFDCAN1->TTIR = FDCAN_TTIR_IWT;
			ok = 0;
		}
		if (FDCAN1->CCCR & FDCAN_CCCR_MON) {
			/* We must restart cycle */
			uint32_t ttir = TTFDCAN1->TTIR;
			TTFDCAN1->TTIR = ttir;
			if (mstate == 0) { /* MON state */
				if (!masterSilentMode && (HAL_GetTick() - stick) > 30) {
					HWFDCAN_Master_Restart(0);
					stick = master_tick = HAL_GetTick();
					mstate = 1;
				}
			} else { /* Non MON state */
				if ((HAL_GetTick() - stick) > 5) {
					SET_BIT(FDCAN1->CCCR, FDCAN_CCCR_INIT);
					while ((FDCAN1->CCCR & FDCAN_CCCR_INIT) == 0U);
					SET_BIT(FDCAN1->CCCR, FDCAN_CCCR_CCE);
					osDelay(10);
					MODIFY_REG(FDCAN1->CCCR, FDCAN_CCCR_MON | FDCAN_CCCR_ASM, FDCAN_CCCR_MON);
					CLEAR_BIT(FDCAN1->CCCR, FDCAN_CCCR_INIT);
					stick = HAL_GetTick();
					mstate = 0;
				} else {
					HWFDCAN_Master_Restart(0);
					master_tick = HAL_GetTick();
				}
			}
		} else {
			/* mstate must be 1 */
			if (ok) {
				mstate = 1;
			} else if((HAL_GetTick() - stick) > 7) {
				SET_BIT(FDCAN1->CCCR, FDCAN_CCCR_INIT);
				while ((FDCAN1->CCCR & FDCAN_CCCR_INIT) == 0U);
				SET_BIT(FDCAN1->CCCR, FDCAN_CCCR_CCE);
				osDelay(10);
				MODIFY_REG(FDCAN1->CCCR, FDCAN_CCCR_MON | FDCAN_CCCR_ASM, FDCAN_CCCR_MON);
				CLEAR_BIT(FDCAN1->CCCR, FDCAN_CCCR_INIT);
				stick = HAL_GetTick();
				mstate = 0;
			}
		}

		/* HAL_FDCAN_EnableTxBufferRequest(&hfdcan1, FT_CAN_TX_CALIBRATION); */
		FDCAN1->TXBAR = FT_CAN_TX_CALIBRATION;

		int skip;

		/* Process input packages */
		do {
			skip = 1;
			/*if (HAL_FDCAN_GetRxFifoFillLevel(&hfdcan1, FDCAN_RX_FIFO1) >= 1) {*/
			if (FDCAN1->RXF1S & FDCAN_RXF1S_F1FL) {
				/* Read discovery message */
				HAL_FDCAN_GetRxMessage(&hfdcan1, FDCAN_RX_FIFO1, &CanRxHeader,
						CanRxData);
				ProcessDiscoveryFromSlave(CanRxData);
				ftstat->canrx++;
				skip = 0;
#if USE_TTCAN_COLLECTOR
		icollector = (icollector + 1) & (TTCAN_COLLECTOR_MAX - 1);
		ttcollector[icollector].tick = HAL_GetTick();
		ttcollector[icollector].ttir = TTFDCAN1->TTIR;
		ttcollector[icollector].cccr = FDCAN1->CCCR;
		ttcollector[icollector].fifo = 1;
		ttcollector[icollector].masterSilentMode = masterSilentMode;
		ttcollector[icollector].masterSilentTick = masterSilentTick;
#endif
			}
			/* if (HAL_FDCAN_GetRxFifoFillLevel(&hfdcan1, FDCAN_RX_FIFO0) >= 1) { */
			if (FDCAN1->RXF0S & FDCAN_RXF0S_F0FL) {
				HAL_FDCAN_GetRxMessage(&hfdcan1, FDCAN_RX_FIFO0, &CanRxHeader,
						CanRxData);
				if (CanRxHeader.Identifier == FT_CAN_MARKER_ID) {
					if (mstate == 0) {
						stick = HAL_GetTick();
						masterSilentMode = 1;
						masterSilentTick = HAL_GetTick();
						continue;
					}
					if (CanRxHeader.DataLength >= FDCAN_DLC_BYTES_8) {
						uint32_t ctick = HAL_GetTick();
						uint32_t ltick = ctick - master_tick;
						uint32_t rtick = (CanRxData[7] << 24) | (CanRxData[6] << 16)
								| (CanRxData[5] << 8) | CanRxData[4];
						if (ltick < rtick) {
							masterSilentMode = 1;
							masterSilentTick = ctick;
							HWFDCAN_Master_Restart(FDCAN_CCCR_MON);
							stick = HAL_GetTick();
							mstate = 0;
							continue;
						}
					}
					/* Something wrong */
					continue;
				}
				ProcessPacketFromSlave(&CanRxHeader, CanRxData);
				ftstat->canrx++;
				skip = 0;
#if USE_TTCAN_COLLECTOR
		icollector = (icollector + 1) & (TTCAN_COLLECTOR_MAX - 1);
		ttcollector[icollector].tick = HAL_GetTick();
		ttcollector[icollector].ttir = TTFDCAN1->TTIR;
		ttcollector[icollector].cccr = FDCAN1->CCCR;
		ttcollector[icollector].fifo = 0;
		ttcollector[icollector].masterSilentMode = masterSilentMode;
		ttcollector[icollector].masterSilentTick = masterSilentTick;
#endif
			}
			if (HAL_FDCAN_GetTxEvent(&hfdcan1, &txevent) == HAL_OK) {
				uint32_t id = txevent.Identifier;

				/* 0 - marker, 1 - calibration, 2 - saddr, config, 3 - op */
				if (id > 1 && (id & 1)) {
					/* We can start next group */
					int n = id >> 1;
					struct ft_shm_txt *txt = getTXT(n);
					int id2 = id & (-2);

					switch(txt->state) {
					case ftslave_state_addr:
						if ((FDCAN1->TXBRP & (1 << id2))) {
							if ((HAL_GetTick() - slavePacketTime[id2])
									> ftcalib.ct) {
								/* Cancel request */
								FDCAN1->TXBCR = 1 << id2;
								ftstat->tdrop++;
							}
							break;
						}
						SendAddrPacket(txt, &TxCommand.command);
						break;
					case ftslave_state_ready:
						if ((FDCAN1->TXBRP & (3 << id2))) {
							uint32_t t = HAL_GetTick();
							if ((FDCAN1->TXBRP & (1 << id2))) {
								if ((t - slavePacketTime[id2]) > ftcalib.ct) {
									FDCAN1->TXBCR = 1 << id2;
									ftstat->tdrop++;
								}
							}
							if ((FDCAN1->TXBRP & (2 << id2))) {
								if ((t - slavePacketTime[id2 + 1]) > ftcalib.ct) {
									FDCAN1->TXBCR = 2 << id2;
									ftstat->tdrop++;
								}
							}
							break;
						}
						SendConfigPacket(n, &TxCommand.command);
						SendOpPacket(txt, &TxCommand.command);
						break;
					default:
						break;
					}
				}
				skip = 0;
			}
		} while(!skip);

		if (!masterSilentMode) {
			ProcessMasterTxPacket(&TxCommand.command);
		}

#if USE_TTCAN_COLLECTOR
		icollector = (icollector + 1) & (TTCAN_COLLECTOR_MAX - 1);
		ttcollector[icollector].tick = HAL_GetTick();
		ttcollector[icollector].ttir = TTFDCAN1->TTIR;
		ttcollector[icollector].cccr = FDCAN1->CCCR;
		ttcollector[icollector].fifo = 4;
		ttcollector[icollector].masterSilentMode = masterSilentMode;
		ttcollector[icollector].masterSilentTick = masterSilentTick;
#endif
		HWFDCAN_Enable_Interrupts();
	}
}

/**
 * @brief Main Slave filter initialization procedure
 * base + 1 for discovery
 * base + even special for input
 * 0x555 calibration
 */
static void HWFDCAN_ConfigureSlaveFilter()
{
	/* No extended id messages */
	sFilterConfig.IdType = FDCAN_EXTENDED_ID;
	sFilterConfig.FilterIndex = 0;
	sFilterConfig.FilterType = FDCAN_FILTER_MASK;
	sFilterConfig.FilterConfig = FDCAN_FILTER_REJECT;
	sFilterConfig.FilterID1 = 0;
	sFilterConfig.FilterID2 = 0;
	sFilterConfig.IsCalibrationMsg = 0;

	HAL_FDCAN_ConfigFilter(&hfdcan1, &sFilterConfig);

	/* Standard id messages */
	sFilterConfig.IdType = FDCAN_STANDARD_ID;
	sFilterConfig.FilterIndex = 0;
	sFilterConfig.FilterType = FDCAN_FILTER_DUAL;
	sFilterConfig.FilterConfig = FDCAN_FILTER_TO_RXFIFO0;
	sFilterConfig.FilterID1 = FT_CAN_DISCOVERY_ID;
	sFilterConfig.FilterID2 = FT_CAN_DISCOVERY_ID;
	sFilterConfig.IsCalibrationMsg = 0;

	HAL_FDCAN_ConfigFilter(&hfdcan1, &sFilterConfig);

	sFilterConfig.IdType = FDCAN_STANDARD_ID;
	sFilterConfig.FilterIndex++;
	sFilterConfig.FilterType = FDCAN_FILTER_DUAL;
	sFilterConfig.FilterConfig = FDCAN_FILTER_TO_RXBUFFER;
	sFilterConfig.FilterID1 = 0x555;
	sFilterConfig.FilterID2 = 0x555;
	sFilterConfig.IsCalibrationMsg = 1;
	sFilterConfig.RxBufferIndex = 0;

	HAL_FDCAN_ConfigFilter(&hfdcan1, &sFilterConfig);

	/* Reject all other packets */
	sFilterConfig.FilterIndex++;
	sFilterConfig.FilterType = FDCAN_FILTER_MASK;
	sFilterConfig.FilterConfig = FDCAN_FILTER_REJECT;
	sFilterConfig.FilterID1 = 0;
	sFilterConfig.FilterID2 = 0;
	sFilterConfig.IsCalibrationMsg = 0;

	HAL_FDCAN_ConfigFilter(&hfdcan1, &sFilterConfig);

	sFilterConfig.IsCalibrationMsg = 0;
}

/**
 * @brief Main Slave procedure for Exteral Engine
 */
static void HWFDCAN_Slave()
{
	CLEAR_BIT(FDCAN1->CCCR, FDCAN_CCCR_DAR);
	HWFDCAN_ConfigureSlaveFilter();
	HWFDCAN_TTCAN_slave();
	TTFDCAN1->TTIR = 0x7ffff;
	HAL_FDCAN_Start(&hfdcan1);

	HWFDCAN_Enable_Interrupts();

	while(1) {
		xSemaphoreTake(hwFDEventHandle, 10);

		/* Check state, restart if required */
		while (TTFDCAN1->TTIR & FDCAN_TTIR_IWT) {
			TTFDCAN1->TTIR = FDCAN_TTIR_IWT;
			osDelay(1);
		}
		if ((TTFDCAN1->TTIR & FDCAN_TTIR_WT)
				|| (FDCAN1->CCCR & FDCAN_CCCR_MON)) {
			/* We must restart cycle */
			SET_BIT(FDCAN1->CCCR, FDCAN_CCCR_INIT);
			uint32_t ttir = TTFDCAN1->TTIR;
			TTFDCAN1->TTIR = ttir;
			while ((FDCAN1->CCCR & FDCAN_CCCR_INIT) == 0U)
				;
			SET_BIT(FDCAN1->CCCR, FDCAN_CCCR_CCE);
			CLEAR_BIT(FDCAN1->CCCR, FDCAN_CCCR_MON);
			CLEAR_BIT(FDCAN1->CCCR, FDCAN_CCCR_CCE);
			CLEAR_BIT(FDCAN1->CCCR, FDCAN_CCCR_INIT);
		}

		int skip;
		do {
			skip = 1;
			if (HAL_FDCAN_GetTxEvent(&hfdcan1, &txevent) == HAL_OK) {
				ProcessSlaveTxPacket(&TxCommand.command);
				skip = 0;
			}
			if (FDCAN1->RXF0S & FDCAN_RXF0S_F0FL) {
				HAL_FDCAN_GetRxMessage(&hfdcan1, FDCAN_RX_FIFO0, &CanRxHeader, CanRxData);
				ProcessPacketFromMaster(CanRxData);
				ftstat->canrx++;
				skip = 0;
			}
			FDCAN1->NDAT1 = FDCAN1->NDAT1; /* Important for RX-filter */
		} while(!skip);

		ProcessSlaveTxPacket(&TxCommand.command);
		HWFDCAN_Enable_Interrupts();
	}
}

/**
 * @brief Main procedure Exteral Master/Slave Engine
 * @param info unused
 */
void HWFDCAN_thread(void *info)
{
	while(!fdcanMode) {
		osDelay(1);
	}

    if (fdcanMode == ft_txt_master) {
    	HWFDCAN_Master();
    } else {
    	HWFDCAN_Slave();
    }
}

/**
  * @brief This function handles FDCAN1 interrupt 0.
  */
void FDCAN1_IT0_IRQHandler(void)
{
	if (hwFDEventHandle != NULL) {
		xSemaphoreGiveFromISR(hwFDEventHandle, NULL);
	}
	FDCAN1->IE = 0;
	TTFDCAN1->TTIE = 0;
	uint32_t ir = FDCAN1->IR;
	FDCAN1->IR = ir;
	ir = TTFDCAN1->TTIR;
	TTFDCAN1->TTIR = ir;
}
